from unittest import mock

import h5py
import numpy as np
from pytest import raises

from .. import VersionedHDF5File
from ..backend import create_base_dataset
from ..hashtable import Hashtable


def test_hashtable(h5file):
    create_base_dataset(h5file, "test_data", data=np.empty((0,)))
    with Hashtable(h5file, "test_data") as h:
        assert len(h) == 0
        h[b"\xff" * 32] = slice(0, 1)
        assert len(h) == 1
        assert h[b"\xff" * 32] == slice(0, 1)
        assert h.largest_index == 1
        assert bytes(h.hash_table[0][0]) == b"\xff" * 32
        assert tuple(h.hash_table[0][1]) == (0, 1)
        assert h == {b"\xff" * 32: slice(0, 1)}

        with raises(TypeError):
            h["\x01" * 32] = slice(0, 1)
        with raises(ValueError):
            h[b"\x01"] = slice(0, 1)
        with raises(TypeError):
            h[b"\x01" * 32] = (0, 1)
        with raises(ValueError):
            h[b"\x01" * 32] = slice(0, 4, 2)


def test_from_raw_data(setup_vfile):
    with setup_vfile("test.h5") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("0") as sv:
            sv.create_dataset("test_data", data=np.arange(100), chunks=(10,))

        h = Hashtable(f, "test_data")
        h_dataset = h.hash_table_dataset
        h2 = Hashtable.from_raw_data(f, "test_data", hash_table_name="test_hash_table")
        h2_dataset = h2.hash_table_dataset
        assert h2_dataset.name == "/_version_data/test_data/test_hash_table"
        np.testing.assert_equal(h_dataset[:], h2_dataset[:])


def test_hashtable_multidimension(h5file):
    # Ensure that the same data with different shape hashes differently
    create_base_dataset(h5file, "test_data", data=np.empty((0,)))
    h = Hashtable(h5file, "test_data")
    assert h.hash(
        np.ones(
            (
                1,
                2,
                3,
            )
        )
    ) != h.hash(np.ones((3, 2, 1)))


def test_issue_208(setup_vfile):
    file = setup_vfile()
    filename = file.filename
    with file as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("0") as sv:
            sv.create_dataset("bar", data=np.arange(10))

    with h5py.File(filename, "r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("1") as sv:
            sv["bar"].resize((12,))
            sv["bar"][8:12] = sv["bar"][6:10]
            sv["bar"][6:8] = [0, 0]


def test_object_dtype_hashes_values(tmp_path):
    """Test that object dtype arrays hash values, not element ids.

    See https://github.com/deshaw/versioned-hdf5/issues/256 for more
    information.
    """
    filename = tmp_path / "test.h5"
    N = 100
    with h5py.File(filename, mode="w") as f:
        file = VersionedHDF5File(f)
        s = ""
        for i in range(N):
            s += "a"
            arr = np.array([s], dtype=object)

            with file.stage_version(f"r{i}") as group:
                group.create_dataset(
                    "values", shape=(1,), dtype=h5py.string_dtype(length=None), data=arr
                )

    with h5py.File(filename, mode="r") as f:
        file = VersionedHDF5File(f)
        for i in range(N):
            assert file[f"r{i}"]["values"][()] == b"a" * (i + 1)


def test_object_dtype_hashes_concatenated_values(tmp_path):
    """Test that object dtype arrays hash values which concatenate
    to the same string to different hashes.
    See https://github.com/deshaw/versioned-hdf5/issues/288.
    """
    filename = tmp_path / "test.h5"
    with h5py.File(filename, mode="w") as f:
        file = VersionedHDF5File(f)
        with file.stage_version("r0") as group:
            group.create_dataset(
                "values",
                dtype=h5py.string_dtype(encoding="ascii"),
                data=np.array([b"a", b"b", b"cd"], dtype=object),
                maxshape=(None,),
                chunks=(100,),
            )
        with file.stage_version("r1") as group:
            group["values"] = np.array([b"ab", b"", b"cd"], dtype=object)
        with file.stage_version("r2") as group:
            group["values"] = np.array([b"ab", b"c", b"d"], dtype=object)

    with h5py.File(filename, mode="r") as f:
        file = VersionedHDF5File(f)
        np.testing.assert_equal(
            file["r0"]["values"][:], np.array([b"a", b"b", b"cd"], dtype=object)
        )
        np.testing.assert_equal(
            file["r1"]["values"][:], np.array([b"ab", b"", b"cd"], dtype=object)
        )
        np.testing.assert_equal(
            file["r2"]["values"][:], np.array([b"ab", b"c", b"d"], dtype=object)
        )


def test_verify_chunk_reuse_data_version_2(tmp_path, monkeypatch):
    """Test whether the issue with DATA_VERSION==2 would have been caught by
    _verify_new_chunk_reuse.

    The problem with the hash function for DATA_VERSION==2 was that it hashed
    the encoded data, not the data itself, meaning that "b'hello'" would hash
    to the same value as b'hello'.
    """

    monkeypatch.setenv("ENABLE_CHUNK_REUSE_VALIDATION", "1")

    def data_version_2_hash(self, data: np.ndarray):
        """
        Compute hash for `data` array.

        (Copied from commit 1f968f4 Hashtable.hash. This version hashes the encoded
        data, not the data itself.)
        """
        if data.dtype == "object":
            hash_value = self.hash_function()
            for value in data.flat:
                hash_value.update(bytes(str(value), "utf-8"))
            hash_value.update(bytes(str(data.shape), "utf-8"))
            return hash_value.digest()
        else:
            return self.hash_function(
                data.data.tobytes() + bytes(str(data.shape), "ascii")
            ).digest()

    with mock.patch.object(Hashtable, "hash", autospec=True) as mocked_hash:
        mocked_hash.side_effect = data_version_2_hash

        data1 = np.array(["b'hello'", "b'world'"], dtype="O")
        data2 = np.array([b"hello", b"world"], dtype="O")

        filename = tmp_path / "data.h5"
        with h5py.File(filename, mode="w") as f:
            vf = VersionedHDF5File(f)
            with vf.stage_version("r0") as group:
                group.create_dataset(
                    "values",
                    dtype=h5py.string_dtype(encoding="ascii"),
                    data=data1,
                    maxshape=(None,),
                    chunks=(2,),
                )

            with raises(ValueError):
                with vf.stage_version("r1") as group:
                    group["values"] = np.concatenate((data2, data2))


def test_verify_chunk_reuse_data_version_3(tmp_path, monkeypatch):
    """Test whether the issue with DATA_VERSION==3 would have been caught by
    _verify_new_chunk_reuse.

    The problem with the hash function for DATA_VERSION==3 was that it didn't
    include the length of string array elements in the hash, meaning that
    the hash of [b"a", b"b", b"cd"] would be the same as [b"ab", b"", b"cd"]
    (because hashing each element in the array is equivalent to hashing
    the concatenated elements of the array).
    """

    monkeypatch.setenv("ENABLE_CHUNK_REUSE_VALIDATION", "1")

    def data_version_3_hash(self, data: np.ndarray):
        """
        Compute hash for `data` array.

        (Copied from commit d382673 Hashtable.hash. This version didn't include the
        string length in the hash.)
        """
        if data.dtype == "object":
            hash_value = self.hash_function()
            for value in data.flat:
                if isinstance(value, str):
                    value = value.encode("utf-8")
                hash_value.update(value)
            hash_value.update(bytes(str(data.shape), "utf-8"))
            return hash_value.digest()
        else:
            return self.hash_function(
                data.data.tobytes() + bytes(str(data.shape), "ascii")
            ).digest()

    with mock.patch.object(Hashtable, "hash", autospec=True) as mocked_hash:
        mocked_hash.side_effect = data_version_3_hash

        filename = tmp_path / "data.h5"
        with h5py.File(filename, mode="w") as f:
            vf = VersionedHDF5File(f)
            with vf.stage_version("r0") as group:
                group.create_dataset(
                    "values",
                    dtype=h5py.string_dtype(encoding="ascii"),
                    data=np.array([b"a", b"b", b"cd"], dtype=object),
                    maxshape=(None,),
                    chunks=(100,),
                )

            with raises(ValueError):
                with vf.stage_version("r1") as group:
                    group["values"] = np.array([b"ab", b"", b"cd"], dtype=object)
                with vf.stage_version("r2") as group:
                    group["values"] = np.array([b"ab", b"c", b"d"], dtype=object)


def test_verify_chunk_reuse_nan(tmp_path):
    """Check that chunks are correctly verified when reused, even with nans."""
    data = np.array([1, 2, 3, 4, 5, np.nan])
    filename = tmp_path / "testdata.h5"
    with h5py.File(filename, mode="w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r0") as sv:
            sv.create_dataset("values", data=data, chunks=(6,))

    with h5py.File(filename, mode="r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r1") as sv:
            sv["values"] = np.concatenate((data, data))


def test_verify_chunk_reuse_strings(tmp_path):
    """Check that that strings can be reused and verified."""
    data = np.array([b"a", b"b", b"cd"], dtype=object)
    filename = tmp_path / "testdata.h5"
    with h5py.File(filename, mode="w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r0") as group:
            group.create_dataset(
                "values",
                data=data,
                dtype=h5py.string_dtype(length=None),
                maxshape=(None,),
                chunks=(3,),
            )

    with h5py.File(filename, mode="r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r1") as group:
            group["values"] = np.concatenate((data, data))


def test_verify_chunk_reuse_multidim_1(tmp_path):
    """Check that we correctly handle chunk reuse verification for multi-dimensional Datasets."""
    filename = tmp_path / "testdata.h5"
    with h5py.File(filename, mode="w") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r0") as group:
            group.create_dataset(
                "values",
                data=np.array([[i + (j % 3) for i in range(8)] for j in range(7)]),
                maxshape=(None, None),
                chunks=(3, 3),
            )

    with h5py.File(filename, mode="r+") as f:
        vf = VersionedHDF5File(f)
        with vf.stage_version("r1") as group:
            values_ds = group["values"]
            values_ds.resize((8, 8))
            values_ds[:] = np.array([[i + (j % 3) for i in range(8)] for j in range(8)])


def test_verify_chunk_disabled_by_default(tmp_path, monkeypatch):
    """Check that we skip chunk reuse verification if the environment variable is not set."""
    monkeypatch.delenv("ENABLE_CHUNK_REUSE_VALIDATION", raising=False)

    # This is the same test as test_verify_chunk_reuse_data_version_2,
    # but with verification turned off with the environment variable.
    def data_version_2_hash(self, data: np.ndarray):
        """
        Compute hash for `data` array.

        (Copied from commit 1f968f4 Hashtable.hash. This version hashes the encoded
        data, not the data itself.)
        """
        if data.dtype == "object":
            hash_value = self.hash_function()
            for value in data.flat:
                hash_value.update(bytes(str(value), "utf-8"))
            hash_value.update(bytes(str(data.shape), "utf-8"))
            return hash_value.digest()
        else:
            return self.hash_function(
                data.data.tobytes() + bytes(str(data.shape), "ascii")
            ).digest()

    with mock.patch.object(Hashtable, "hash", autospec=True) as mocked_hash:
        mocked_hash.side_effect = data_version_2_hash

        data1 = np.array(["b'hello'", "b'world'"], dtype="O")
        data2 = np.array([b"hello", b"world"], dtype="O")

        filename = tmp_path / "data.h5"
        with h5py.File(filename, mode="w") as f:
            vf = VersionedHDF5File(f)
            with vf.stage_version("r0") as group:
                group.create_dataset(
                    "values",
                    dtype=h5py.string_dtype(encoding="ascii"),
                    data=data1,
                    maxshape=(None,),
                    chunks=(2,),
                )

            # This should raise an error, but will not because chunk
            # reuse verification is turned off.
            with vf.stage_version("r1") as group:
                group["values"] = np.concatenate((data2, data2))



import numpy as np
from ndindex import Slice, Tuple
from numpy.testing import assert_equal
from pytest import raises

from ..backend import DEFAULT_CHUNK_SIZE
from ..versions import (
    all_versions,
    commit_version,
    create_version_group,
    delete_version,
    get_nth_previous_version,
    set_current_version,
)


def test_create_version(h5file):
    chunk_size = 2**10
    chunks = (chunk_size,)
    data = np.concatenate(
        (np.ones((2 * chunk_size,)), 2 * np.ones(chunks), 3 * np.ones(chunks))
    )

    version1 = create_version_group(h5file, "version1", "")
    commit_version(
        version1,
        {"test_data": data},
        chunks={"test_data": chunks},
        compression={"test_data": "gzip"},
        compression_opts={"test_data": 3},
    )

    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, chunks={"test_data": (2**9,)}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, compression={"test_data": "lzf"}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, compression_opts={"test_data": 4}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    assert version1.attrs["prev_version"] == "__first_version__"
    assert version1.parent.attrs["current_version"] == "version1"
    # Test against the file here, not version1, since version1 is the
    # InMemoryGroup returned from create_version_group, but we did not add
    # the datasets to it directly.
    assert_equal(h5file["_version_data/versions/version1/test_data"], data)

    ds = h5file["/_version_data/test_data/raw_data"]

    assert ds.shape == (3 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert ds.compression == "gzip"
    assert ds.compression_opts == 3

    data[0] = 0.0
    version2 = create_version_group(h5file, "version2", "version1")

    raises(
        ValueError,
        lambda: commit_version(version1, {"test_data": data}, make_current=False),
    )
    commit_version(version2, {"test_data": data}, make_current=False)
    assert version2.attrs["prev_version"] == "version1"
    assert_equal(h5file["_version_data/versions/version2/test_data"], data)
    assert version2.parent.attrs["current_version"] == "version1"

    assert ds.shape == (4 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert_equal(ds[3 * chunk_size], 0.0)
    assert_equal(ds[3 * chunk_size + 1 : 4 * chunk_size], 1.0)
    assert ds.compression == "gzip"
    assert ds.compression_opts == 3

    assert set(all_versions(h5file)) == {"version1", "version2"}
    assert set(all_versions(h5file, include_first=True)) == {
        "version1",
        "version2",
        "__first_version__",
    }


def test_create_version_chunks(h5file):
    chunk_size = 2**10
    chunks = (chunk_size,)
    data = np.concatenate(
        (np.ones((2 * chunk_size,)), 2 * np.ones(chunks), 3 * np.ones(chunks))
    )
    # TODO: Support creating the initial version with chunks
    version1 = create_version_group(h5file, "version1")
    commit_version(
        version1,
        {"test_data": data},
        chunks={"test_data": chunks},
        compression={"test_data": "gzip"},
        compression_opts={"test_data": 3},
    )
    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, chunks={"test_data": (2**9,)}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, compression={"test_data": "lzf"}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    version_bad = create_version_group(h5file, "version_bad", "")
    raises(
        ValueError,
        lambda: commit_version(
            version_bad, {"test_data": data}, compression_opts={"test_data": 4}
        ),
    )
    delete_version(h5file, "version_bad", "version1")

    assert_equal(h5file["_version_data/versions/version1/test_data"], data)

    ds = h5file["/_version_data/test_data/raw_data"]

    assert ds.shape == (3 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert ds.compression == "gzip"
    assert ds.compression_opts == 3

    data2_chunks = {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): np.ones(chunks),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): np.ones(chunks),
        Tuple(Slice(2 * chunk_size, 3 * chunk_size, 1)): 2 * np.ones(chunks),
        Tuple(Slice(3 * chunk_size, 4 * chunk_size, 1)): 3 * np.ones(chunks),
    }
    data2_chunks[Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1))][0] = 0.0
    data[0] = 0.0

    version2 = create_version_group(h5file, "version2")
    commit_version(version2, {"test_data": data2_chunks})
    assert_equal(h5file["_version_data/versions/version2/test_data"], data)

    assert ds.shape == (4 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert_equal(ds[3 * chunk_size], 0.0)
    assert_equal(ds[3 * chunk_size + 1 : 4 * chunk_size], 1.0)
    assert ds.compression == "gzip"
    assert ds.compression_opts == 3

    data3_chunks = {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): np.ones(chunks),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): Slice(
            0 * chunk_size, 1 * chunk_size
        ),
        Tuple(Slice(2 * chunk_size, 3 * chunk_size, 1)): Slice(
            1 * chunk_size, 2 * chunk_size
        ),
        Tuple(Slice(3 * chunk_size, 4 * chunk_size, 1)): Slice(
            2 * chunk_size, 3 * chunk_size
        ),
    }
    data3_chunks[Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1))][0] = 2.0
    data[0] = 2.0

    version3 = create_version_group(h5file, "version3")
    commit_version(version3, {"test_data": data3_chunks})
    assert_equal(h5file["_version_data/versions/version3/test_data"], data)

    assert ds.shape == (5 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert_equal(ds[3 * chunk_size], 0.0)
    assert_equal(ds[3 * chunk_size + 1 : 4 * chunk_size], 1.0)
    assert_equal(ds[4 * chunk_size], 2.0)
    assert_equal(ds[4 * chunk_size + 1 : 5 * chunk_size], 1.0)

    assert set(all_versions(h5file)) == {"version1", "version2", "version3"}


def test_get_nth_prev_version(h5file):
    data = np.concatenate(
        (
            np.ones((2 * DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )

    version1 = create_version_group(h5file, "version1")
    commit_version(version1, {"test_data": data})

    data[0] = 2.0
    version2 = create_version_group(h5file, "version2")
    commit_version(version2, {"test_data": data})

    data[0] = 3.0
    version3 = create_version_group(h5file, "version3")
    commit_version(version3, {"test_data": data})

    data[1] = 2.0
    version2_1 = create_version_group(h5file, "version2_1", "version1")
    commit_version(version2_1, {"test_data": data})

    assert get_nth_previous_version(h5file, "version1", 0) == "version1"

    with raises(IndexError):
        get_nth_previous_version(h5file, "version1", 1)

    assert get_nth_previous_version(h5file, "version2", 0) == "version2"
    assert get_nth_previous_version(h5file, "version2", 1) == "version1"
    with raises(IndexError):
        get_nth_previous_version(h5file, "version2", 2)

    assert get_nth_previous_version(h5file, "version3", 0) == "version3"
    assert get_nth_previous_version(h5file, "version3", 1) == "version2"
    assert get_nth_previous_version(h5file, "version3", 2) == "version1"
    with raises(IndexError):
        get_nth_previous_version(h5file, "version3", 3)

    assert get_nth_previous_version(h5file, "version2_1", 0) == "version2_1"
    assert get_nth_previous_version(h5file, "version2_1", 1) == "version1"
    with raises(IndexError):
        get_nth_previous_version(h5file, "version2_1", 2)


def test_set_current_version(h5file):
    data = np.concatenate(
        (
            np.ones((2 * DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )

    version1 = create_version_group(h5file, "version1")
    commit_version(version1, {"test_data": data})
    versions = h5file["_version_data/versions"]
    assert versions.attrs["current_version"] == "version1"

    data[0] = 2.0
    version2 = create_version_group(h5file, "version2")
    commit_version(version2, {"test_data": data})
    assert versions.attrs["current_version"] == "version2"

    set_current_version(h5file, "version1")
    assert versions.attrs["current_version"] == "version1"

    with raises(ValueError):
        set_current_version(h5file, "version3")


def test_delete_version(h5file):
    raises(ValueError, lambda: delete_version(h5file, "version1"))

    data = np.concatenate(
        (
            np.ones((2 * DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )

    version1 = create_version_group(h5file, "version1")
    commit_version(version1, {"test_data": data})
    versions = h5file["_version_data/versions"]
    assert versions.attrs["current_version"] == "version1"

    raises(ValueError, lambda: delete_version(h5file, "version1", "doesntexist"))

    delete_version(h5file, "version1")
    versions = h5file["_version_data/versions"]

    assert versions.attrs["current_version"] == "__first_version__"
    assert list(versions) == ["__first_version__"]


def test_forbidden_dataset_name(h5file):
    data = np.concatenate(
        (
            np.ones((2 * DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )
    v1 = create_version_group(h5file, "v1")

    with raises(ValueError):
        commit_version(v1, {"versions": data})



import os

import h5py

from versioned_hdf5 import VersionedHDF5File
from versioned_hdf5.wrappers import InMemoryArrayDataset

try:
    from versioned_hdf5.wrappers import DatasetWrapper
except ImportError:

    class DatasetWrapper:
        pass


import numpy as np


class TimeInMemoryArrayDataset:
    timeout = 1000

    def teardown(self):
        os.remove("bench.hdf5")

    def time_getattr(self):
        with h5py.File("bench.hdf5", "w") as f:
            versioned_file = VersionedHDF5File(f)
            with versioned_file.stage_version("version1") as g:
                dataset = g.create_dataset(
                    "data",
                    data=np.arange(10000).reshape((100, 10, 10)),
                    chunks=(3, 3, 3),
                )
                assert (
                    isinstance(dataset, InMemoryArrayDataset)
                    or isinstance(dataset, DatasetWrapper)
                    and isinstance(dataset.dataset, InMemoryArrayDataset)
                )
                dataset[:, 0, 0:6]

    def time_setattr(self):
        with h5py.File("bench.hdf5", "w") as f:
            versioned_file = VersionedHDF5File(f)
            with versioned_file.stage_version("version1") as g:
                dataset = g.create_dataset(
                    "data",
                    data=np.arange(10000).reshape((100, 10, 10)),
                    chunks=(3, 3, 3),
                )
                assert (
                    isinstance(dataset, InMemoryArrayDataset)
                    or isinstance(dataset, DatasetWrapper)
                    and isinstance(dataset.dataset, InMemoryArrayDataset)
                )
                dataset[:, 0, 0:6] = -1

    def time_resize_bigger(self):
        with h5py.File("bench.hdf5", "w") as f:
            versioned_file = VersionedHDF5File(f)
            with versioned_file.stage_version("version1") as g:
                dataset = g.create_dataset(
                    "data",
                    data=np.arange(10000).reshape((100, 10, 10)),
                    chunks=(3, 3, 3),
                )
                assert (
                    isinstance(dataset, InMemoryArrayDataset)
                    or isinstance(dataset, DatasetWrapper)
                    and isinstance(dataset.dataset, InMemoryArrayDataset)
                )
                dataset.resize((100, 100, 100))

    def time_resize_smaller(self):
        with h5py.File("bench.hdf5", "w") as f:
            versioned_file = VersionedHDF5File(f)
            with versioned_file.stage_version("version1") as g:
                dataset = g.create_dataset(
                    "data",
                    data=np.arange(10000).reshape((100, 10, 10)),
                    chunks=(3, 3, 3),
                )
                assert (
                    isinstance(dataset, InMemoryArrayDataset)
                    or isinstance(dataset, DatasetWrapper)
                    and isinstance(dataset.dataset, InMemoryArrayDataset)
                )
                dataset.resize((10, 10, 10))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# import versioned_hdf5 from compiled and installed extension
# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))


# -- Project information -----------------------------------------------------

project = "Versioned HDF5"
copyright = "2020, Quansight"
author = "Quansight"

# Enable warnings for all bad cross references. These are turned into errors
# with the -W flag in the Makefile.
nitpicky = True

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx_multiversion",
]

# sphinx-multiversion configuration
smv_tag_whitelist = r"^[0-9].[0-9].*$"
smv_branch_whitelist = "master"
smv_remote_whitelist = None
smv_released_pattern = r"^tags/.*$"
smv_outputdir_format = "{ref.name}"
smv_prefer_remote_refs = False

graphviz_output_format = "svg"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_type_aliases = {"File": "h5py.File"}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

html_theme_options = {
    "github_user": "deshaw",
    "github_repo": "versioned-hdf5",
    "github_banner": False,  # https://github.com/bitprophet/alabaster/issues/166
    "github_button": False,
    # 'logo_name': True,
    "travis_button": False,
    "show_related": True,
    # Remove gray background from inline code
    "code_bg": "#FFFFFF",
    # Originally 940px
    "page_width": "1000px",
    # Fonts
    "font_family": "Palatino, 'goudy old style', 'minion pro', 'bell mt', Georgia, 'Hiragino Mincho Pro', serif",
    "font_size": "18px",
    "code_font_family": "'Menlo', 'DejaVu Sans Mono', 'Consolas', 'Bitstream Vera Sans Mono', monospace",
    "code_font_size": "0.85em",
}

html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
        "versioning.html",
    ]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Lets us use single backticks for code
default_role = "code"



import os
import shutil
import tempfile

import h5py
import numpy

from versioned_hdf5 import VersionedHDF5File

filename = "delete_versions_bench.h5"


try:
    from versioned_hdf5 import delete_versions
except ImportError:
    from versioned_hdf5.replay import recreate_dataset, swap, tmp_group

    def delete_versions(f, versions_to_delete, names=("values",)):
        """
        Modified replay.delete_version to delete multiple versions.
        """
        if isinstance(f, VersionedHDF5File):
            f = f.f

        def callback(dataset, version_name):
            if version_name in versions_to_delete:
                return
            return dataset

        newf = tmp_group(f)

        for name in names:
            recreate_dataset(f, name, newf, callback=callback)

        swap(f, newf)

        for version in versions_to_delete:
            del f["_version_data/versions"][version]

        del newf[newf.name]


class TimeDeleting:
    params = [10, 30, 50]
    timeout = 1000

    def setup(self, n):
        if not os.path.exists(filename):
            with h5py.File(filename, "w") as f:
                vf = VersionedHDF5File(f)
                with vf.stage_version("init") as sv:
                    sv.create_dataset(
                        "values",
                        shape=(0, 0),
                        dtype="float",
                        fillvalue=numpy.nan,
                        chunks=(22, 100),
                        maxshape=(None, None),
                        compression="lzf",
                    )

            # generate some test data with around 1000 versions
            v = 1
            with h5py.File(filename, "r+") as f:
                vf = VersionedHDF5File(f)
                for d in range(3):
                    with vf.stage_version(str(v)) as sv:
                        values_ds = sv["values"]
                        values_ds.resize(
                            (values_ds.shape[0] + 1, values_ds.shape[1] + 5000)
                        )
                        values_ds[-1, -5000] = numpy.random.rand()
                        v += 1
                    for c in range(n):
                        with vf.stage_version(str(v)) as sv:
                            values_ds = sv["values"]
                            idxs = numpy.random.choice(
                                values_ds.shape[1], 50, replace=False
                            )
                            values_ds[-1, idxs] = numpy.random.rand(50)
                            v += 1

    def teardown(self, n):
        os.remove(filename)

    def time_delete(self, n):
        tmp_name = tempfile.mktemp(".h5")
        shutil.copy2(filename, tmp_name)
        try:
            # want to keep only every 10th version
            versions_to_delete = []
            with h5py.File(tmp_name, "r") as f:
                vf = VersionedHDF5File(f)
                versions = sorted(
                    [(v, vf._versions[v].attrs["timestamp"]) for v in vf._versions],
                    key=lambda t: t[1],
                )
                for i, v in enumerate(versions):
                    if i % 10 != 0:
                        versions_to_delete.append(v[0])

            with h5py.File(tmp_name, "r+") as f:
                delete_versions(f, versions_to_delete)
        finally:
            os.remove(tmp_name)



# distutils: language=c++
import sys
from functools import lru_cache

import cython
import h5py
from h5py import h5d

cimport numpy as np

import numpy as np
from cython import void
from h5py import h5s
from h5py._hl.base import phil
from h5py.h5t import py_create
from ndindex import Slice, Tuple
from numpy.typing import ArrayLike

from libc.stddef cimport ptrdiff_t, size_t
from libc.stdio cimport FILE, fclose
from libcpp.vector cimport vector

from versioned_hdf5.cytools import np_hsize_t
from versioned_hdf5.cytools cimport ceil_a_over_b, count2stop, hsize_t, stop2count
from versioned_hdf5.tools import asarray


cdef FILE* fmemopen(void* buf, size_t size, const char* mode):
    raise NotImplementedError("fmemopen is not available on Windows")
if sys.platform != "win32":
    from posix.stdio cimport fmemopen

ctypedef ptrdiff_t ssize_t

cdef extern from "hdf5.h":
    # HDF5 types copied from h5py/api_types_hdf5.pxd
    ctypedef long int hid_t
    ctypedef int hbool_t
    ctypedef int herr_t
    ctypedef int htri_t

    cdef herr_t H5Dread(
        hid_t dset_id,
        hid_t mem_type_id,
        hid_t mem_space_id,
        hid_t file_space_id,
        hid_t dxpl_id,
        void* buf,
    ) nogil

    cdef hid_t H5E_DEFAULT = 0
    cdef herr_t H5Eprint(hid_t stack_id, FILE* stream) nogil

    cdef hid_t H5I_INVALID_HID = -1

    cdef hid_t H5P_DEFAULT = 0
    # virtual Dataset functions
    cdef hid_t H5Pget_virtual_vspace(hid_t dcpl_id, size_t index) nogil
    cdef hid_t H5Pget_virtual_srcspace(hid_t dcpl_id, size_t index) nogil

    ctypedef enum H5S_sel_type:
        H5S_SEL_ERROR = -1,  # Error
        H5S_SEL_NONE = 0,  # Nothing selected
        H5S_SEL_POINTS = 1,  # Sequence of points selected
        H5S_SEL_HYPERSLABS = 2,  # "New-style" hyperslab selection defined
        H5S_SEL_ALL = 3,  # Entire extent selected
        H5S_SEL_N = 4  # THIS MUST BE LAST

    ctypedef enum H5S_seloper_t:
        H5S_SELECT_NOOP = -1,
        H5S_SELECT_SET = 0,
        H5S_SELECT_OR,
        H5S_SELECT_AND,
        H5S_SELECT_XOR,
        H5S_SELECT_NOTB,
        H5S_SELECT_NOTA,
        H5S_SELECT_APPEND,
        H5S_SELECT_PREPEND,
        H5S_SELECT_INVALID    # Must be the last one

    cdef H5S_sel_type H5Sget_select_type(
        hid_t space_id
    ) except H5S_sel_type.H5S_SEL_ERROR nogil
    cdef herr_t H5Sclose(hid_t space_id) nogil
    cdef int H5Sget_simple_extent_ndims(hid_t space_id) nogil
    cdef hid_t H5Screate_simple	(
        int rank,
        const hsize_t* dims,
        const hsize_t* maxdims,
    ) nogil
    cdef htri_t H5Sget_regular_hyperslab(
        hid_t spaceid,
        hsize_t* start,
        hsize_t* stride,
        hsize_t* count,
        hsize_t* block,
    ) nogil
    cdef herr_t H5Sselect_hyperslab(
        hid_t space_id,
        H5S_seloper_t op,
        const hsize_t* start,
        const hsize_t* stride,
        const hsize_t* count,
        const hsize_t* block,
    ) nogil

# https://github.com/numpy/numpy/blob/b0a52b76c2bebe5c237722589d13bf02affa9c43/numpy/core/include/numpy/ndarraytypes.h#L30
cdef enum:
    NPY_MAXDIMS = 32

np.import_array()

NP_GE_200 = np.lib.NumpyVersion(np.__version__) >= "2.0.0"


def spaceid_to_slice(space) -> Tuple:
    """
    Convert an h5py spaceid object into an ndindex index

    The resulting index is always a Tuple index.
    """

    sel_type = space.get_select_type()

    if sel_type == h5s.SEL_ALL:
        return Tuple()
    elif sel_type == h5s.SEL_HYPERSLABS:
        slices = []
        starts, strides, counts, blocks = space.get_regular_hyperslab()
        for start, stride, count, block in zip(starts, strides, counts, blocks):
            slices.append(hyperslab_to_slice(start, stride, count, block))
        return Tuple(*slices)
    elif sel_type == h5s.SEL_NONE:
        return Tuple(Slice(0, 0))
    else:
        raise NotImplementedError("Point selections are not yet supported")


@lru_cache(2048)
def hyperslab_to_slice(start, stride, count, block):
    if not (block == 1 or count == 1):
        raise NotImplementedError("Nontrivial blocks are not yet supported")
    end = start + (stride * (count - 1) + 1) * block
    stride = stride if block == 1 else 1
    return Slice(start, end, stride)


@cython.infer_types(True)
cdef _spaceid_to_slice(space_id: hid_t):
    """
    Helper function to read the data for `space_id` selection and
    convert it to a Tuple of slices.
    """
    sel_type = H5Sget_select_type(space_id)

    if sel_type == H5S_sel_type.H5S_SEL_ALL:
        return Tuple()
    elif sel_type == H5S_sel_type.H5S_SEL_HYPERSLABS:
        slices = []

        rank = H5Sget_simple_extent_ndims(space_id)
        if rank < 0:
            raise HDF5Error()
        start_array = vector[hsize_t](rank)
        stride_array = vector[hsize_t](rank)
        count_array = vector[hsize_t](rank)
        block_array = vector[hsize_t](rank)

        if H5Sget_regular_hyperslab(
            space_id,
            start_array.data(),
            stride_array.data(),
            count_array.data(),
            block_array.data(),
        ) < 0:
            raise HDF5Error()

        for i in range(rank):
            start = start_array[i]
            stride = stride_array[i]
            count = count_array[i]
            block = block_array[i]
            if not (block == 1 or count == 1):
                raise NotImplementedError("Nontrivial blocks are not yet supported")
            end = start + (stride * (count - 1) + 1) * block
            stride = stride if block == 1 else 1
            slices.append(Slice(start, end, stride))

        return Tuple(*slices)
    elif sel_type == H5S_sel_type.H5S_SEL_NONE:
        return Tuple(Slice(0, 0))
    else:
        raise NotImplementedError("Point selections are not yet supported")


@cython.infer_types(True)
cpdef build_data_dict(dcpl, raw_data_name: str):
    """
    Function to build the "data_dict" of a versioned virtual dataset.

    All virtual datasets created by versioned-hdf5 should have chunks in
    exactly one raw dataset `raw_data_name` in the same file.
    This function blindly assumes this is the case.

    :param dcpl: the dataset creation property list of the versioned dataset
    :param raw_data_name: the name of the corresponding raw dataset
    :return: a dictionary mapping the `Tuple` of the virtual dataset chunk
        to a `Slice` in the raw dataset.
    """
    data_dict = {}

    with phil:
        dcpl_id: hid_t = dcpl.id
        virtual_count: size_t = dcpl.get_virtual_count()

        for j in range(virtual_count):
            vspace_id = H5Pget_virtual_vspace(dcpl_id, j)
            if vspace_id == H5I_INVALID_HID:
                raise HDF5Error()
            try:
                vspace_slice_tuple = _spaceid_to_slice(vspace_id)
            finally:
                if H5Sclose(vspace_id) < 0:
                    raise HDF5Error()

            srcspace_id = H5Pget_virtual_srcspace(dcpl_id, j)
            if srcspace_id == H5I_INVALID_HID:
                raise HDF5Error()
            try:
                srcspace_slice_tuple = _spaceid_to_slice(srcspace_id)
            finally:
                if H5Sclose(srcspace_id) < 0:
                    raise HDF5Error()

            # the slice into the raw_data (srcspace_slice_tuple) is only
            # on the first axis
            data_dict[vspace_slice_tuple] = srcspace_slice_tuple.args[0]

    return data_dict


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.infer_types(True)
cpdef tuple[np.ndarray, np.ndarray] build_slab_indices_and_offsets(
    dcpl,
    shape: tuple[int, ...],
    chunk_size: tuple[int, ...],
):
    """Build the slab_indices and slab_offsets arrays for a StagedChangesArray that
    accesses a versioned virtual dataset.

    :param dcpl: the dataset creation property list of the versioned dataset
    :return: tuple of (slab_indices, slab_offsets) arrays, each with the same shape as
    the chunks in the virtual dataset.
    """
    cdef hsize_t[NPY_MAXDIMS] chunk_size_arr
    cdef hsize_t[NPY_MAXDIMS] nchunks_arr
    cdef hsize_t[NPY_MAXDIMS] h5_start
    cdef hsize_t[NPY_MAXDIMS] scratch  # Dumping ground for unused outputs

    nchunks = []
    ndim = len(shape)
    for j in range(ndim):
        chunk_size_j: hsize_t = chunk_size[j]
        assert chunk_size_j > 0
        chunk_size_arr[j] = chunk_size[j]
        nchunks_j = ceil_a_over_b(shape[j], chunk_size_j)
        nchunks_arr[j] = nchunks_j
        nchunks.append(nchunks_j)

    # Default for missing chunks is to point to slabs[0],
    # which is a single chunk covered in fill_value
    cdef np.ndarray slab_indices = np.zeros(nchunks, dtype=np_hsize_t)
    cdef np.ndarray slab_offsets = np.zeros(nchunks, dtype=np_hsize_t)
    indices_strides = slab_indices.strides
    slab_indices_data = slab_indices.data
    slab_offsets_data = slab_offsets.data

    with phil:
        # Newly created sparse datasets completely full of fill_value are not virtual.
        # FIXME Error getting virtual sources on empty virtual dataset
        # https://github.com/h5py/h5py/issues/1660
        if dcpl.get_layout() != h5d.VIRTUAL or 0 in shape:
            return slab_indices, slab_offsets

        dcpl_id: hid_t = dcpl.id
        virtual_count: size_t = dcpl.get_virtual_count()

        with nogil:
            # Iterate over all chunks except full ones
            for chunk_i in range(virtual_count):
                # Calculate offset of the chunk, IN BYTES, relative to the start of the
                # flattened buffer of slab_indices and slab_offsets
                vspace_id = H5Pget_virtual_vspace(dcpl_id, chunk_i)
                if vspace_id == H5I_INVALID_HID:
                    raise HDF5Error()
                try:
                    vspace_type = H5Sget_select_type(vspace_id)
                    if vspace_type != H5S_sel_type.H5S_SEL_HYPERSLABS:
                        raise ValueError(f"Unexpected {vspace_type=}")

                    # Get chunk index in the virtual dataset
                    if H5Sget_regular_hyperslab(
                        vspace_id,
                        h5_start,
                        scratch,  # IGNORE - stride (always 1)
                        scratch,  # IGNORE - count
                        scratch,  # IGNORE - block (typically, but not always, 1)
                    ) < 0:
                        raise HDF5Error()
                finally:
                    if H5Sclose(vspace_id) < 0:
                        raise HDF5Error()

                indices_offset: ssize_t = 0
                for j in range(ndim):
                    chunk_idx = h5_start[j] // chunk_size_arr[j]
                    if h5_start[j] % chunk_size_arr[j] > 0:
                        raise ValueError("Misaligned chunk")
                    if chunk_idx >= nchunks_arr[j]:
                        raise ValueError("chunk_idx too high")
                    indices_offset += chunk_idx * indices_strides[j]

                # Calculate offset of the chunk, IN POINTS,
                # relative to the start of raw_data along axis 0
                srcspace_id = H5Pget_virtual_srcspace(dcpl_id, chunk_i)
                if srcspace_id == H5I_INVALID_HID:
                    raise HDF5Error()
                try:
                    srcspace_type = H5Sget_select_type(srcspace_id)
                    if srcspace_type != H5S_sel_type.H5S_SEL_HYPERSLABS:
                        raise ValueError(f"Unexpected {srcspace_type=}")

                    # Get coordinates in raw_data
                    if H5Sget_regular_hyperslab(
                        srcspace_id,
                        h5_start,  # Axis 0 = slab_offset; other axes = always 0
                        scratch,   # IGNORE - stride (always 1)
                        scratch,   # IGNORE - count
                        scratch,   # IGNORE - block (typically, but not always, 1)
                    ) < 0:
                        raise HDF5Error()
                finally:
                    if H5Sclose(srcspace_id) < 0:
                        raise HDF5Error()

                slab_offset = h5_start[0]

                # Write single point into slab_indices and slab_offsets
                # NOTE: indices_offset is in bytes on the flat buffer;
                #       slab_offset is in points on the rows.
                # slabs[0] = fill_value; slabs[1] = raw_data.
                # See StagedChangesArray initialisation in wrappers.py.
                (<hsize_t*>(slab_indices_data + indices_offset))[0] = 1
                (<hsize_t*>(slab_offsets_data + indices_offset))[0] = slab_offset

    return slab_indices, slab_offsets


cdef Exception HDF5Error():
    """Generate a RuntimeError with the HDF5 error message.

    This function must be invoked only after a HDF5 function returned an error code.
    """
    if sys.platform == "win32":
        # No fmemopen available
        return RuntimeError("HDF5 error")

    cdef char buf[20000]
    cdef FILE* stream = fmemopen(buf, sizeof(buf), "w")
    if stream == NULL:
        return MemoryError("fmemopen() failed")
    with phil:
        H5Eprint(H5E_DEFAULT, stream)
    fclose(stream)
    msg = buf.decode("utf-8", errors="replace")
    return RuntimeError(msg)


cpdef void read_many_slices(
    src: np.ndarray | h5py.Dataset,
    np.ndarray dst,
    src_start: ArrayLike,
    dst_start: ArrayLike,
    count: ArrayLike,
    src_stride: ArrayLike | None = None,
    dst_stride: ArrayLike | None = None,
    fast: bool | None = None,
):
    """Transfer multiple n-dimensional slices of data from src to dst::

        dst[dst_idx0] = src[src_idx0]
        dst[dst_idx1] = src[src_idx1]
        ...

    where all indices are tuples of slices.

    Parameters
    ----------
    src: np.ndarray | h5py.Dataset
        The source data to read from
    dst: np.ndarray
        The destination array to write to.
        It must be C-contiguous and of the same dtype and dimensionality as src.
    src_start: array-like
        The starting coordinates, a.k.a. offsets, of the slices in src.
        It must be a 2D array or array-like (e.g. list of lists) with the same number of
        rows as there are selections and as many columns as there are dimensions in src.
    dst_start: array-like
        The starting coordinates of the slices in dst.
        Same format as src_start.
    count: array-like
        The number of elements to read and write in each dimension.
        Same format as src_start.
    src_stride: array-like, optional
        The stride, a.k.a. step, of the slices when reading from src, in points.
        Note that this differs from numpy.ndarray.strides, which is in bytes.
        Same format as src_start. If omitted, default to 1.
    dst_stride: array-like, optional
        The stride of the slices when writing to dst.
        Same format as src_start. If omitted, default to 1.
    fast: bool, optional
        If True, use the hdf5 C API directly to read the data if possible. This is when
        src is a h5py dataset with a simple data type (see h5py.Dataset._fast_read_ok).
        If False, use pure Python numpy syntax to slice src and dst.
        If omitted, determine automatically. It's recommended to omit this flag unless
        you're running a unit test or benchmark.

    For example, if src and dst are 2D and strides are omitted or always 1,
    the areas of data to be copied will be::

        dst[
            dst_start[i, 0]:dst_start[i, 0] + count[i, 0],
            dst_start[i, 1]:dst_start[i, 1] + count[i, 1],
        ] = src[
            src_start[i, 0]:src_start[i, 0] + count[i, 0],
            src_start[i, 1]:src_start[i, 1] + count[i, 1],
        ]

    where i is the row of each of the coordinates arrays.

    Coordinates arrays can alternatively be 1D arrays, with one point per dimension;
    they will be broadcasted against the other arrays.
    For example, you may use a 1D array for count if you want all selected areas to have
    the same shape.

    Scalar selections
    -----------------
    If dst has less dimensions than src, you should present a reshaped view of dst
    to this function.

    e.g. to execute::

        dst[2:3, 4:6] = src[6:7, 8, 9:11]

    where dst.ndim == 2 and src.ndim == 3, you should call::

        read_many_slices(
            src,
            dst[:, None, :],
            src_start=[[6, 8, 9]],
            dst_start=[[2, 0, 4]],
            count=[[1, 1, 2]],
        )

    Note that the same hack is not possible when src is a h5py.Dataset and has less
    dimensions than dst, as src[:, None, :] would load the whole dataset into memory.

    TODO support src.ndim < dst.ndim. It would take some extra logic
    and an extra parameter to inform where the extra dimensions are.

    Performance notes
    -----------------
    When reading from h5py, this function calls::

        H5Sselect_hyperslab(file_id, H5S_SELECT_SET, ...)
        H5Sselect_hyperslab(mem_id, H5S_SELECT_SET, ...)
        H5DRead(...)

    once per input row; see _read_many_slices_fast.

    This is much faster than calling H5Sselect_hyperslab(..., H5S_SELECT_OR, ...) once
    per input row followed by a single call to H5DRead, a.k.a. hyperslab fusion,
    due to a known issue in libhdf5:
    https://forum.hdfgroup.org/t/union-of-non-consecutive-hyperslabs-is-very-slow/5062

    The downside of this approach is that if your selections access the same chunk more
    than once, they have to be contiguous or else the chunk may fall out of the
    internal cache of libhdf5. This can also happen with contiguous selections that span
    multiple chunks.
    """
    # Begin input validation and preprocessing

    ndim = dst.ndim
    if ndim < 1:
        raise ValueError("must operate on at least one dimension")
    if dst.dtype != src.dtype or ndim != src.ndim or not dst.flags.writeable:
        raise ValueError(
            "dst must be a writeable numpy array of the same dtype and dimensionality "
            "as src"
        )
    if dst.size == 0 or not all(src.shape):
        return

    cdef bint bfast = False
    if fast is not False:
        if isinstance(src, h5py.Dataset):
            with phil:
                bfast = src._fast_read_ok
        if fast and not bfast:
            raise ValueError("fast transfer is not possible with this source")

    if bfast and not dst.flags.c_contiguous:
        # TODO we could support non-C-contiguous arrays by doing
        # arithmetics with starts and strides
        raise NotImplementedError("dst must be C-contiguous for fast mode")

    src_start = _preproc_many_slices_idx(src_start, ndim, bfast)
    dst_start = _preproc_many_slices_idx(dst_start, ndim, bfast)
    count = _preproc_many_slices_idx(count, ndim, bfast)

    if src_stride is None:
        src_stride = np.ones(ndim, dtype=np_hsize_t)
    else:
        src_stride = _preproc_many_slices_idx(src_stride, ndim, bfast)
    if dst_stride is None:
        dst_stride = np.ones(ndim, dtype=np_hsize_t)
    else:
        dst_stride = _preproc_many_slices_idx(dst_stride, ndim, bfast)

    for arr in (src_start, dst_start, count, src_stride, dst_stride):
        if arr.size == 0:
            return

    # dummy makes sure to coerce into the correct number of dimensions when
    # all indices are 1D arrays.
    # This raises if the arrays have mismatched number of rows or dimensions, or if
    # there's a wrong number of columns...
    dummy = np.empty((1, ndim))
    _, src_start, dst_start, count, src_stride, dst_stride = np.broadcast_arrays(
        dummy, src_start, dst_start, count, src_stride, dst_stride
    )
    # ...*unless* ndim=1, in which case we need to test explicitly
    if src_start.shape[1] != ndim:
        raise ValueError("Coordinates arrays must have as many columns as src.ndim")

    cdef np.ndarray src_shape = np.array(src.shape, dtype=np_hsize_t)
    cdef np.npy_intp[1] ndim_ptr = {ndim}
    cdef np.ndarray dst_shape = np.PyArray_SimpleNewFromData(
        1, ndim_ptr, np.NPY_INTP, <void*>dst.shape
    )
    # On 32-bit platforms, sizeof(hsize_t) == 8; sizeof(npy_intp) == 4
    # On 64-bit platforms, don't copy unnecessarily
    dst_shape = asarray(dst_shape, np_hsize_t)

    clipped_count = _clip_count(
        src_shape,
        dst_shape,
        src_start,
        dst_start,
        count,
        src_stride,
        dst_stride,
    )

    # End of input validation and preprocessing

    if bfast:
        _read_many_slices_fast(
            src,
            dst,
            <hsize_t*>dst_shape.data,
            src_start,
            dst_start,
            clipped_count,
            src_stride,
            dst_stride,
        )
    else:
        _read_many_slices_slow(
            src,
            dst,
            src_start,
            dst_start,
            clipped_count,
            src_stride,
            dst_stride,
        )


cdef np.ndarray _preproc_many_slices_idx(obj: ArrayLike, hsize_t ndim, bint fast):
    """Helper of read_many_slices. Ensure that obj is a numpy array of hsize_t with
    either 1 or 2 dimensions, and that it's C-contiguous at least along the
    innermost dimension.
    """
    # np.asarray with unsigned dtype raises in numpy>=2 if presented with a Python list
    # containing negative numbers, but can quietly cause an integer underflow if arr is
    # already a numpy array with signed dtype and negative numbers in it.
    # TODO https://github.com/numpy/numpy/issues/25396
    if not NP_GE_200:
        obj = np.asarray(obj)
    if hasattr(obj, "dtype") and obj.dtype.kind != "u" and (obj < 0).any():
        raise OverflowError("index out of bounds for uint64")
    # Don't copy when converting from np.intp to uint64 on 64-bit platforms
    cdef np.ndarray arr = asarray(obj, np_hsize_t)

    if arr.ndim not in (1, 2):
        raise ValueError("Coordinates arrays must have 1 or 2 dimensions")

    # The array must be contiguous along the columns in order to be passed to
    # H5Sselect_hyperslab. So a view like a[:2] is OK, but a[::2] must be made
    # contiguous first. Note that, to save a bit of time, this deep-copy happens before
    # we broadcast 1D arrays to add rows.
    if fast and arr.strides[arr.ndim - 1] != sizeof(hsize_t):
        return np.ascontiguousarray(arr)
    else:
        return arr


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.infer_types(True)
cdef hsize_t[:, :] _clip_count(
    const hsize_t[:] src_shape,
    const hsize_t[:] dst_shape,
    const hsize_t[:, :] src_start,
    const hsize_t[:, :] dst_start,
    const hsize_t[:, :] count,
    const hsize_t[:, :] src_stride,
    const hsize_t[:, :] dst_stride,
):
    """Helper of read_many_slices. Clip count to prevent writing out of bounds.
    This function also neuters src_start beyond src.shape or dst_start beyond dst.shape.
    Finally, it also ensures that strides are >= 1.

    TODO refactor as a @cython.ufunc.
    Cython ufuncs don't support hsize_t (long long) at the time of writing.
    """
    nslices, ndim = count.shape[:2]
    cdef hsize_t[:, :] out = np.empty((nslices, ndim), dtype=np_hsize_t)

    with nogil:
        for i in range(nslices):
            for j in range(ndim):
                if src_stride[i, j] == 0 or dst_stride[i, j] == 0:
                    raise ValueError("Strides must be strictly greater than zero")

                out[i, j] = min(
                    count[i, j],
                    stop2count(src_start[i, j], src_shape[j], src_stride[i, j]),
                    stop2count(dst_start[i, j], dst_shape[j], dst_stride[i, j]),
                )

    return out


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.infer_types(True)
cdef void _read_many_slices_fast (
    src: h5py.Dataset,
    np.ndarray dst,
    const hsize_t* dst_shape,
    const hsize_t[:, :] src_start,
    const hsize_t[:, :] dst_start,
    const hsize_t[:, :] count,
    const hsize_t[:, :] src_stride,
    const hsize_t[:, :] dst_stride,
):
    """Implements read_many_slices data transfer when src is a h5py.Dataset with simple
    extents only (h5py.Dataset._fast_read_ok).

    Performance notes
    -----------------
    See docstring of read_many_slices
    """
    nslices, ndim = src_start.shape[:2]

    with phil:
        # h5py internals. See h5py/dataset.py and h5py/_selector.pyx.
        dset_id: hid_t = src.id.id
        # must remain in scope for the id to remain valid
        space = src.id.get_space()
        file_space_id: hid_t = space.id
        # must remain in scope for the id to remain valid
        mem_type = py_create(src.dtype)
        mem_type_id: hid_t = mem_type.id

        # libhdf5 is not thread safe (read comment about locking in h5py/_objects.pyx),
        # so we are wrapping everything in a reentrant lock (`with phil`).
        # We can however release the GIL to allow unrelated code to run in parallel.
        with nogil:
            mem_space_id = H5Screate_simple(ndim, dst_shape, NULL)
            if mem_space_id == H5I_INVALID_HID:
                raise HDF5Error()

            try:
                for i in range(nslices):
                    for j in range(ndim):
                        if count[i, j] == 0:
                            break
                    else:
                        # count > 0 along all axes
                        if H5Sselect_hyperslab(
                            file_space_id,
                            H5S_SELECT_SET,
                            &src_start[i, 0],
                            &src_stride[i, 0],
                            &count[i, 0],
                            NULL,
                        ) < 0:
                            raise HDF5Error()

                        if H5Sselect_hyperslab(
                            mem_space_id,
                            H5S_SELECT_SET,
                            &dst_start[i, 0],
                            &dst_stride[i, 0],
                            &count[i, 0],
                            NULL,
                        ) < 0:
                            raise HDF5Error()

                        if H5Dread(
                            dset_id,
                            mem_type_id,
                            mem_space_id,
                            file_space_id,
                            H5P_DEFAULT,
                            dst.data,
                        ) < 0:
                            raise HDF5Error()
            finally:
                if H5Sclose(mem_space_id) < 0:
                    raise HDF5Error()


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.infer_types(True)
cdef void _read_many_slices_slow (
    src: np.ndarray | h5py.Dataset,
    np.ndarray dst,
    const hsize_t[:, :] src_start,
    const hsize_t[:, :] dst_start,
    const hsize_t[:, :] count,
    const hsize_t[:, :] src_stride,
    const hsize_t[:, :] dst_stride,
):
    """Implements read_many_slices data transfer when fast transfer cannot be performed.

    This happens when:
    1. src is a h5py.Dataset but h5py.Dataset._fast_read_ok returns False.
       This is to avoid replicating the complex machinery found in
       h5py.Dataset.__getitem__.
    2. src is a numpy array.

    Performance notes
    -----------------
    There has been experimentation to bypass the Python slicing machinery by hacking the
    numpy C API. While functionally successful, there was no material performance gain.

    An alternative approach would be to manually loop over the strides at least for
    basic item sizes (1, 2, 4, and 8 bytes). This has not been attempted yet.
    """
    nslices, ndim = src_start.shape[:2]

    for i in range(nslices):
        for j in range(ndim):
            if count[i, j] == 0:
                break
        else:
            # count > 0 along all axes
            src_idx = []
            dst_idx = []
            for j in range(ndim):
                src_start_ij = src_start[i, j]
                dst_start_ij = dst_start[i, j]
                count_ij = count[i, j]
                src_stride_ij = src_stride[i, j]
                dst_stride_ij = dst_stride[i, j]

                src_stop_ij = count2stop(src_start_ij, count_ij, src_stride_ij)
                dst_stop_ij = count2stop(dst_start_ij, count_ij, dst_stride_ij)

                src_idx.append(slice(src_start_ij, src_stop_ij, src_stride_ij))
                dst_idx.append(slice(dst_start_ij, dst_stop_ij, dst_stride_ij))

            dst[tuple(dst_idx)] = src[tuple(src_idx)]



from __future__ import annotations

import itertools
from typing import Any

import hypothesis
import numpy as np
import pytest
from hypothesis import given
from hypothesis import strategies as st
from numpy.testing import assert_array_equal

from ..cytools import np_hsize_t
from ..slicetools import read_many_slices
from ..subchunk_map import (
    DROP_AXIS,
    EntireChunksMapper,
    SliceMapper,
    TransferType,
    as_subchunk_map,
    index_chunk_mappers,
    read_many_slices_params_nd,
)

max_examples = 10_000


def non_negative_step_slices_st(size: int):
    start = st.one_of(st.none(), st.integers(-size - 1, size + 1))
    stop = st.one_of(st.none(), st.integers(-size - 1, size + 1))
    # only non-negative steps (or None) are allowed
    step = st.one_of(st.none(), st.integers(1, size + 1))
    return st.builds(slice, start, stop, step)


def array_indices_st(size: int):
    return st.one_of(
        st.lists(st.integers(-size, max(0, size - 1)), max_size=size * 2),
        st.lists(st.booleans(), min_size=size, max_size=size),
    )


def scalar_indices_st(size: int):
    return st.integers(-size, size - 1) if size > 0 else st.nothing()


@st.composite
def basic_idx_st(draw, shape: tuple[int, ...]):
    """Hypothesis draw of slice and integer indexes"""
    nidx = draw(st.integers(0, len(shape)))
    idx_st = st.tuples(
        *(
            st.one_of(non_negative_step_slices_st(size), scalar_indices_st(size))
            for size in shape[:nidx]
        )
    )
    return draw(idx_st)


@st.composite
def fancy_idx_st(draw, shape: tuple[int, ...]) -> Any:
    """A single axis is indexed by either

    - a list[int] whose elements can be negative, non-unique, and not in order, or
    - a list[bool]

    All other axes are indexed by slices.

    Interleaving scalars and slices and array indices is not supported:
    https://github.com/Quansight-Labs/ndindex/issues/188
    """
    fancy_idx_axis = draw(st.integers(0, len(shape) - 1))
    nidx = draw(st.integers(fancy_idx_axis + 1, len(shape)))
    idx_st = st.tuples(
        *[non_negative_step_slices_st(shape[dim]) for dim in range(fancy_idx_axis)],
        array_indices_st(shape[fancy_idx_axis]),
        *[
            non_negative_step_slices_st(shape[dim])
            for dim in range(fancy_idx_axis + 1, nidx)
        ],
    )
    return draw(idx_st)


@st.composite
def shape_chunks_st(
    draw, max_ndim: int = 4, min_size: int = 1, max_size: int = 20
) -> tuple[tuple[int, ...], tuple[int, ...]]:
    shape_st = st.lists(st.integers(min_size, max_size), min_size=1, max_size=max_ndim)
    shape = tuple(draw(shape_st))

    chunks_st = st.tuples(*[st.integers(1, s + 1) for s in shape])
    chunks = draw(chunks_st)
    return shape, chunks


def idx_st(shape: tuple[int, ...]) -> Any:
    return st.one_of(basic_idx_st(shape), fancy_idx_st(shape))


@st.composite
def idx_shape_chunks_st(
    draw, max_ndim: int = 4
) -> tuple[Any, tuple[int, ...], tuple[int, ...]]:
    shape, chunks = draw(shape_chunks_st(max_ndim))
    idx = draw(idx_st(shape))
    return idx, shape, chunks


@pytest.mark.slow
@given(idx_shape_chunks_st())
@hypothesis.settings(max_examples=max_examples, deadline=None)
def test_as_subchunk_map(args):
    idx, shape, chunks = args

    source = np.arange(1, np.prod(shape) + 1, dtype=np.int32).reshape(shape)
    expect = source[idx]
    actual = np.zeros_like(expect)

    for chunk_idx, value_sub_idx, chunk_sub_idx in as_subchunk_map(idx, shape, chunks):
        chunk_idx = chunk_idx.raw

        # Test that chunk_idx selects whole chunks
        assert isinstance(chunk_idx, tuple)
        assert len(chunk_idx) == len(chunks)
        for i, c, d in zip(chunk_idx, chunks, shape):
            assert isinstance(i, slice)
            assert i.start % c == 0
            assert i.stop == min(i.start + c, d)
            assert i.step == 1

        assert not actual[value_sub_idx].any(), "overlapping value_sub_idx"
        actual[value_sub_idx] = source[chunk_idx][chunk_sub_idx]

    assert_array_equal(actual, expect)


def test_mapper_attributes():
    _, (mapper,) = index_chunk_mappers(slice(5), (6,), (3,))
    assert mapper.dset_size == 6
    assert mapper.chunk_size == 3
    assert mapper.n_chunks == 2
    assert mapper.last_chunk_size == 3

    _, (mapper,) = index_chunk_mappers((), (5,), (3,))
    assert mapper.dset_size == 5
    assert mapper.chunk_size == 3
    assert mapper.n_chunks == 2
    assert mapper.last_chunk_size == 2


@pytest.mark.slow
@given(idx_shape_chunks_st(max_ndim=1))
@hypothesis.settings(max_examples=max_examples, deadline=None)
def test_chunks_indexer(args):
    """Test IndexChunkMapper.chunks_indexer and IndexChunkMapper.whole_chunks_idxidx"""
    idx, shape, chunks = args
    _, mappers = index_chunk_mappers(idx, shape, chunks)
    if not mappers:
        return  # Early exit for empty index
    assert len(shape) == len(chunks) == len(mappers) == 1
    dset_size = shape[0]
    mapper = mappers[0]
    assert mapper.dset_size == shape[0]
    assert mapper.chunk_size == chunks[0]

    source = np.arange(1, dset_size + 1)
    expect = source[idx]
    actual = np.zeros_like(expect)

    all_chunks = np.arange(mapper.n_chunks)
    sel_chunks = all_chunks[mapper.chunks_indexer()]
    whole_chunks = sel_chunks[mapper.whole_chunks_idxidx()]

    # Test that the slices of chunks are strictly monotonic ascending
    assert_array_equal(sel_chunks, np.unique(sel_chunks))
    assert_array_equal(whole_chunks, np.unique(whole_chunks))

    # Test that whole_chunks is a subset of sel_chunks
    assert np.setdiff1d(whole_chunks, sel_chunks, assume_unique=True).size == 0

    for i in sel_chunks:
        source_idx, value_sub_idx, chunk_sub_idx = mapper.chunk_submap(i)
        chunk = source[source_idx.raw]

        if value_sub_idx is DROP_AXIS:
            value_sub_idx = ()
        actual[value_sub_idx] = chunk[chunk_sub_idx]

        coverage = np.zeros_like(chunk)
        coverage[chunk_sub_idx] = 1
        assert coverage.any(), "chunk selected by chunk_indexer() is not covered"
        if i in whole_chunks:
            assert coverage.all(), "whole chunk is partially covered"
        else:
            assert not coverage.all(), "partial chunk is wholly covered"

    assert_array_equal(actual, expect)


@pytest.mark.slow
@given(idx_shape_chunks_st(max_ndim=1))
@hypothesis.settings(max_examples=max_examples, deadline=None)
def test_read_many_slices_param(args):
    idx, shape, chunks = args
    _, mappers = index_chunk_mappers(idx, shape, chunks)
    if not mappers:
        return  # Early exit for empty index
    assert len(shape) == len(chunks) == len(mappers) == 1
    dset_size = shape[0]
    chunk_size = chunks[0]
    mapper = mappers[0]
    assert mapper.dset_size == shape[0]
    assert mapper.chunk_size == chunks[0]

    source = np.arange(1, dset_size + 1, dtype=np.int32)
    expect = source[idx]
    actual = np.zeros_like(expect)

    slab_offsets = np.arange(0, dset_size, chunk_size, dtype=np_hsize_t)
    assert len(slab_offsets) == mapper.n_chunks

    # Randomize the order of the chunks on the slab.
    # Note that the last chunk may be smaller than chunk_size,
    # so we may end up with an area of the slab full off zeros.
    slab_offsets = np.random.permutation(slab_offsets)
    slab = np.zeros(mapper.n_chunks * chunk_size, dtype=source.dtype)
    for i, offset in enumerate(slab_offsets.tolist()):
        chunk = source[i * chunk_size : (i + 1) * chunk_size]
        slab[offset : offset + len(chunk)] = chunk

    n_sel_chunks = len(mapper.chunk_indices)
    slab_offsets = slab_offsets[mapper.chunks_indexer()]
    assert len(slab_offsets) == n_sel_chunks

    slices, chunks_to_slices = mapper.read_many_slices_params()
    if chunks_to_slices is None:
        assert slices.shape == (n_sel_chunks, 5)
    else:
        # Test that we are not generating a 1:N mapping in the special case of a fancy
        # index that contains exactly one point per chunk, and therefore
        # read_many_slices() needs to transfer exactly one slice per chunk;
        # e.g. chunk_size=10, idx=[3, 15, 21].
        assert slices.shape[0] > n_sel_chunks
        assert slices.shape[1] == 5

        assert chunks_to_slices.shape == (n_sel_chunks + 1,)
        assert chunks_to_slices[0] == 0
        assert chunks_to_slices[-1] == len(slices)

        # Replicate slab_offsets for each slice of each chunk
        slab_offsets_idx = []
        for i in range(n_sel_chunks):
            start = chunks_to_slices[i]
            stop = chunks_to_slices[i + 1]
            assert 0 <= start < stop <= len(slices)
            count = stop - start
            slab_offsets_idx += [i] * count
        slab_offsets = slab_offsets[slab_offsets_idx]
    assert len(slab_offsets) == len(slices)

    actual_view = actual[None] if actual.shape == () else actual

    read_many_slices(
        src=slab,
        dst=actual_view,
        src_start=slices[:, 0:1] + slab_offsets[:, None],
        dst_start=slices[:, 1:2],
        count=slices[:, 2:3],
        src_stride=slices[:, 3:4],
        dst_stride=slices[:, 4:5],
    )

    assert_array_equal(actual, expect)


@pytest.mark.slow
@given(idx_shape_chunks_st())
@hypothesis.settings(max_examples=max_examples, deadline=None)
def test_read_many_slices_param_nd(args):
    idx, shape, chunks = args
    _, mappers = index_chunk_mappers(idx, shape, chunks)
    if not mappers:
        return  # Early exit for empty index

    source = np.arange(1, np.prod(shape) + 1, dtype=np.int32).reshape(shape)
    expect = source[idx]

    chunk_idxidx = np.array(
        list(itertools.product(*[list(range(len(m.chunk_indices))) for m in mappers])),
        dtype=np_hsize_t,
    )

    # Generate the slab and the slab_offsets.
    # We don't care to populate the slab with the full contents of the source dataset,
    # we just need the chunks impacted by the index.
    n_chunks = np.prod([m.n_chunks for m in mappers])
    slab = np.zeros((n_chunks * chunks[0], *chunks[1:]), dtype=source.dtype)
    slab_offsets = np.random.choice(  # randomize slab order
        np.arange(0, n_chunks * chunks[0], chunks[0], dtype=np_hsize_t),
        size=len(chunk_idxidx),
        replace=False,
    )
    for chunk_idxidx_i, slab_offset_i in zip(chunk_idxidx, slab_offsets.tolist()):
        source_idx = []
        for j, chunk_idxidx_ij in enumerate(chunk_idxidx_i):
            chunk_idx_ij = int(mappers[j].chunk_indices[chunk_idxidx_ij])
            source_idx.append(
                slice(start := chunk_idx_ij * chunks[j], start + chunks[j])
            )
        chunk = source[tuple(source_idx)]
        # chunk may be smaller than the chunk size along the edges
        slab[slab_offset_i:][tuple(slice(c) for c in chunk.shape)] = chunk

    # Can't pass "hsize_t[:] | None" to cythonized functions
    DUMMY_SLAB_OFFSETS = np.empty(0, dtype=np_hsize_t)

    getitem_dst = np.zeros_like(expect)
    getitem_dst_view = getitem_dst[tuple(m.value_view_idx for m in mappers)]

    # __getitem__
    getitem_slices_nd = read_many_slices_params_nd(
        TransferType.getitem,
        mappers,
        chunk_idxidx,
        src_slab_offsets=slab_offsets,
        dst_slab_offsets=DUMMY_SLAB_OFFSETS,
    )
    read_many_slices(
        src=slab,
        dst=getitem_dst_view,
        src_start=getitem_slices_nd[:, 0, :],
        dst_start=getitem_slices_nd[:, 1, :],
        count=getitem_slices_nd[:, 2, :],
        src_stride=getitem_slices_nd[:, 3, :],
        dst_stride=getitem_slices_nd[:, 4, :],
    )

    assert_array_equal(getitem_dst, expect)

    # Now do it again in the other way for __setitem__
    setitem_dst_slab = np.zeros_like(slab)
    setitem_slices_nd = read_many_slices_params_nd(
        TransferType.setitem,
        mappers,
        chunk_idxidx,
        src_slab_offsets=DUMMY_SLAB_OFFSETS,
        dst_slab_offsets=slab_offsets,
    )
    read_many_slices(
        src=getitem_dst_view,
        dst=setitem_dst_slab,
        src_start=setitem_slices_nd[:, 0, :],
        dst_start=setitem_slices_nd[:, 1, :],
        count=setitem_slices_nd[:, 2, :],
        src_stride=setitem_slices_nd[:, 3, :],
        dst_stride=setitem_slices_nd[:, 4, :],
    )
    getitem_dst2 = np.zeros_like(getitem_dst)
    getitem_dst_view2 = getitem_dst2[tuple(m.value_view_idx for m in mappers)]
    read_many_slices(
        src=setitem_dst_slab,
        dst=getitem_dst_view2,
        src_start=getitem_slices_nd[:, 0, :],
        dst_start=getitem_slices_nd[:, 1, :],
        count=getitem_slices_nd[:, 2, :],
        src_stride=getitem_slices_nd[:, 3, :],
        dst_stride=getitem_slices_nd[:, 4, :],
    )
    assert_array_equal(getitem_dst2, expect)

    # And finally slab-to-slab
    slab2slab_dst = np.zeros_like(slab)
    dst_slab_offsets = np.random.permutation(slab_offsets)
    slab2slab_slices_nd = read_many_slices_params_nd(
        TransferType.slab_to_slab,
        mappers,
        chunk_idxidx,
        src_slab_offsets=slab_offsets,
        dst_slab_offsets=dst_slab_offsets,
    )
    read_many_slices(
        src=slab,
        dst=slab2slab_dst,
        src_start=slab2slab_slices_nd[:, 0, :],
        dst_start=slab2slab_slices_nd[:, 1, :],
        count=slab2slab_slices_nd[:, 2, :],
        src_stride=slab2slab_slices_nd[:, 3, :],
        dst_stride=slab2slab_slices_nd[:, 4, :],
    )
    # In order to test the new slab, read from it with __getitem__
    getitem_slices_nd3 = read_many_slices_params_nd(
        TransferType.getitem,
        mappers,
        chunk_idxidx,
        src_slab_offsets=dst_slab_offsets,
        dst_slab_offsets=DUMMY_SLAB_OFFSETS,
    )
    getitem_dst3 = np.zeros_like(getitem_dst)
    getitem_dst_view3 = getitem_dst3[tuple(m.value_view_idx for m in mappers)]
    read_many_slices(
        src=slab2slab_dst,
        dst=getitem_dst_view3,
        src_start=getitem_slices_nd3[:, 0, :],
        dst_start=getitem_slices_nd3[:, 1, :],
        count=getitem_slices_nd3[:, 2, :],
        src_stride=getitem_slices_nd3[:, 3, :],
        dst_stride=getitem_slices_nd3[:, 4, :],
    )
    assert_array_equal(getitem_dst3, expect)


@pytest.mark.slow
@given(idx_shape_chunks_st(max_ndim=1))
@hypothesis.settings(max_examples=max_examples, deadline=None)
def test_entire_chunks_mapper(args):
    idx, shape, chunks = args
    _, mappers = index_chunk_mappers(idx, shape, chunks)
    if not mappers:
        return  # Early exit for empty index
    assert len(shape) == len(chunks) == len(mappers) == 1
    chunk_size = chunks[0]
    orig_mapper = mappers[0]
    entire_mapper = EntireChunksMapper(orig_mapper)

    # Test chunks_indexer() and whole_chunks_idxidx()
    all_chunks = np.arange(orig_mapper.n_chunks)
    sel_chunks = all_chunks[orig_mapper.chunks_indexer()]
    np.testing.assert_array_equal(
        all_chunks[entire_mapper.chunks_indexer()], sel_chunks
    )
    np.testing.assert_array_equal(
        sel_chunks[entire_mapper.whole_chunks_idxidx()], sel_chunks
    )

    # Test read_many_slices_params()
    entire_slices, entire_chunks_to_slices = entire_mapper.read_many_slices_params()
    assert entire_chunks_to_slices is None

    n_sel_chunks = len(orig_mapper.chunk_indices)
    expect_count = np.full(n_sel_chunks, chunk_size, dtype=np_hsize_t)
    if orig_mapper.chunk_indices[-1] == orig_mapper.n_chunks - 1:
        expect_count[-1] = orig_mapper.last_chunk_size

    assert_array_equal(
        entire_slices,
        np.stack(
            [
                np.zeros(n_sel_chunks, dtype=np_hsize_t),  # src_start,
                np.asarray(orig_mapper.chunk_indices) * chunk_size,  # dst_start,
                expect_count,  # count
                np.ones(n_sel_chunks, dtype=np_hsize_t),  # src_stride,
                np.ones(n_sel_chunks, dtype=np_hsize_t),  # dst_stride,
            ],
            axis=1,
        ),
        strict=True,
    )


def test_simplify_indices():
    """Test that a fancy index that can be redefined globally as a slice results in a
    SliceMapper
    """
    _, (mapper,) = index_chunk_mappers([True, True, False, False], (4,), (2,))
    assert isinstance(mapper, SliceMapper)
    assert mapper.start == 0
    assert mapper.stop == 2
    assert mapper.step == 1

    _, (mapper,) = index_chunk_mappers([False, True, False, True], (4,), (2,))
    assert isinstance(mapper, SliceMapper)
    assert mapper.start == 1
    assert mapper.stop == 4
    assert mapper.step == 2


def test_chunk_submap_simplifies_indices():
    """Test that, when a fancy index can't be globally simplified to a slice,
    as_subchunk_map still attemps to simplify the individual chunk subindices.
    """
    _, (mapper,) = index_chunk_mappers(
        [True, False, True, False]  # chunk 0
        + [True, True, False, False]  # chunk 1
        + [True, False, True, True],  # chunk 2
        (12,),
        (4,),
    )
    _, value_sub_idx, chunk_sub_idx = mapper.chunk_submap(0)
    assert value_sub_idx == slice(0, 2, 1)
    assert chunk_sub_idx == slice(0, 3, 2)
    _, value_sub_idx, chunk_sub_idx = mapper.chunk_submap(1)
    assert value_sub_idx == slice(2, 4, 1)
    assert chunk_sub_idx == slice(0, 2, 1)
    _, value_sub_idx, chunk_sub_idx = mapper.chunk_submap(2)
    assert value_sub_idx == slice(4, 7, 1)
    assert_array_equal(chunk_sub_idx, [0, 2, 3])  # Can't be simplified


def test_chunks_indexer_simplifies_indices():
    """Test that chunks_indexer() and whole_chunks_indexer() return a slice if possible"""
    _, (mapper,) = index_chunk_mappers(slice(None, None, 3), (10,), (2,))
    assert_array_equal(mapper.chunks_indexer(), [0, 1, 3, 4])

    _, (mapper,) = index_chunk_mappers(slice(None, None, 4), (10,), (2,))
    assert mapper.chunks_indexer() == slice(0, 5, 2)

    _, (mapper,) = index_chunk_mappers(
        [True, False, True, True, False, False], (6,), (2,)
    )
    assert mapper.chunks_indexer() == slice(0, 2, 1)
    assert mapper.whole_chunks_idxidx() == slice(1, 2, 1)

    _, (mapper,) = index_chunk_mappers(
        [True, False, True, True, False, False, True, True, True, True], (10,), (2,)
    )
    assert_array_equal(mapper.chunks_indexer(), [0, 1, 3, 4])
    assert_array_equal(mapper.whole_chunks_idxidx(), slice(1, 4, 1))

    _, (mapper,) = index_chunk_mappers(
        [True, False, False, True, True, False],
        (6,),
        (2,),
    )
    assert mapper.chunks_indexer() == slice(0, 3, 1)
    assert mapper.whole_chunks_idxidx() == slice(0, 0, 1)


def test_invalid_indices():
    with pytest.raises(IndexError, match="too many indices"):
        index_chunk_mappers((0, 0), (4,), (2,))
    with pytest.raises(IndexError, match="out of bounds"):
        index_chunk_mappers((4,), (4,), (2,))
    with pytest.raises(IndexError, match="out of bounds"):
        index_chunk_mappers((-5,), (4,), (2,))
    with pytest.raises(IndexError, match="out of bounds"):
        index_chunk_mappers(([4],), (4,), (2,))
    with pytest.raises(IndexError, match="out of bounds"):
        index_chunk_mappers(([-5],), (4,), (2,))
    with pytest.raises(IndexError, match="boolean index did not match indexed array"):
        index_chunk_mappers(([True] * 3,), (4,), (2,))
    with pytest.raises(IndexError, match="boolean index did not match indexed array"):
        index_chunk_mappers(([True] * 5,), (4,), (2,))

    with pytest.raises(IndexError, match="valid indices"):
        index_chunk_mappers("foo", (4,), (2,))
    with pytest.raises(ValueError, match="step"):
        index_chunk_mappers(slice(None, None, 0), (4,), (2,))

    with pytest.raises(NotImplementedError, match="step"):
        index_chunk_mappers(slice(None, None, -1), (4,), (2,))
    with pytest.raises(NotImplementedError, match="None"):
        index_chunk_mappers(None, (4,), (2,))
    with pytest.raises(NotImplementedError, match="newaxis"):
        index_chunk_mappers(np.newaxis, (4,), (2,))
    with pytest.raises(NotImplementedError, match="Ellipsis"):
        index_chunk_mappers((..., 0), (4, 4), (2, 2))

    # Fancy indices are tentatively simplified to slices. However, it would be a
    # mistake to simplify [[0, 1], [0, 1]] to [:2, :2]!
    with pytest.raises(NotImplementedError, match="Multiple fancy indices"):
        index_chunk_mappers(([0, 1], [0, 1]), (4, 4), (2, 2))
    with pytest.raises(NotImplementedError, match="Multiple fancy indices"):
        index_chunk_mappers(([True, True], [True, True]), (2, 2), (2, 2))
    with pytest.raises(NotImplementedError, match="Multiple fancy indices"):
        index_chunk_mappers(([True, True], [0, 1]), (2, 2), (2, 2))
    with pytest.raises(NotImplementedError, match="1-dimensional"):
        index_chunk_mappers([[0, 1], [1, 0]], (4,), (2,))

    with pytest.raises(ValueError, match="chunk sizes"):
        index_chunk_mappers((), (4,), (0,))
    with pytest.raises(ValueError, match="shape"):
        index_chunk_mappers((), (-1,), (2,))
    with pytest.raises(ValueError, match="chunks"):
        index_chunk_mappers((), (4,), (2, 2))



from __future__ import annotations

import cython

if cython.compiled:  # pragma: nocover
    # This is repeated here from the header in order to silence mypy, which now treats
    # hsize_t as Any, without breaking cython.
    from cython.cimports.versioned_hdf5.cytools import hsize_t

# numpy equivalent dtype for hsize_t
from numpy import uint64 as np_hsize_t  # noqa: F401


@cython.ccall
@cython.nogil
@cython.exceptval(check=False)
def stop2count(start: hsize_t, stop: hsize_t, step: hsize_t) -> hsize_t:
    """Given a start:stop:step slice or range, return the number of elements yielded.

    This is functionally identical to::

        len(range(start, stop, step))

    Doesn't assume that stop >= start. Assumes that step >= 1.
    """
    # Note that hsize_t is unsigned so stop - start could underflow.
    if stop <= start:
        return 0
    return (stop - start - 1) // step + 1


@cython.ccall
@cython.nogil
@cython.exceptval(check=False)
def count2stop(start: hsize_t, count: hsize_t, step: hsize_t) -> hsize_t:
    """Inverse of stop2count.

    When count == 0 or when step>1, multiple stops can yield the same count.
    This function returns the smallest stop >= start.
    """
    if count == 0:
        return start
    return start + (count - 1) * step + 1


@cython.ccall
@cython.nogil
@cython.exceptval(check=False)
def ceil_a_over_b(a: hsize_t, b: hsize_t) -> hsize_t:
    """Returns ceil(a/b). Assumes a >= 0 and b > 0.

    Note
    ----
    This module is compiled with the cython.cdivision flag and the data type is
    unsigned. This causes behaviour to change if one accidentally passes negative a or b
    when running in pure python mode.
    """
    return a // b + (a % b > 0)


@cython.ccall
@cython.nogil
@cython.exceptval(check=False)
def smallest_step_after(x: hsize_t, a: hsize_t, m: hsize_t) -> hsize_t:
    """Find the smallest integer y >= x where y = a + k*m for whole k's
    Assumes 0 <= a <= x and m >= 1.

    a                  x    y
    | <-- m --> | <-- m --> |
    """
    return a + ceil_a_over_b(x - a, m) * m



import os

import h5py

from versioned_hdf5 import VersionedHDF5File
from versioned_hdf5.wrappers import InMemoryDataset

try:
    from versioned_hdf5.wrappers import DatasetWrapper
except ImportError:

    class DatasetWrapper:
        pass


import numpy as np


class TimeInMemoryDataset:
    timeout = 1000

    def setup(self):
        if hasattr(self, "file"):
            self.file.close()
        if os.path.exists("bench.hdf5"):
            os.remove("bench.hdf5")

        with h5py.File("bench.hdf5", "w") as f:
            versioned_file = VersionedHDF5File(f)

            with versioned_file.stage_version("version1") as g:
                g.create_dataset(
                    "data",
                    data=np.arange(10000).reshape((100, 10, 10)),
                    chunks=(3, 3, 3),
                )

        self.file = h5py.File("bench.hdf5", "a")
        self.versioned_file = VersionedHDF5File(self.file)

    # def teardown(self):
    #     self.file.close()
    #     os.remove('bench.hdf5')

    def time_getitem(self):
        dataset = self.versioned_file["version1"]["data"]
        assert (
            isinstance(dataset, InMemoryDataset)
            or isinstance(dataset, DatasetWrapper)
            and isinstance(dataset.dataset, InMemoryDataset)
        )
        dataset[:, 0, 0:6]

    def time_setitem(self):
        # https://github.com/airspeed-velocity/asv/issues/966
        self.setup()
        with self.versioned_file.stage_version("version2") as g:
            dataset = g["data"]
            assert (
                isinstance(dataset, InMemoryDataset)
                or isinstance(dataset, DatasetWrapper)
                and isinstance(dataset.dataset, InMemoryDataset)
            )
            dataset[:, 0, 0:6] = -1

    def time_resize_bigger(self):
        # https://github.com/airspeed-velocity/asv/issues/966
        self.setup()
        with self.versioned_file.stage_version("version2") as g:
            dataset = g["data"]
            assert (
                isinstance(dataset, InMemoryDataset)
                or isinstance(dataset, DatasetWrapper)
                and isinstance(dataset.dataset, InMemoryDataset)
            )
            dataset.resize((100, 100, 100))

    def time_resize_smaller(self):
        # https://github.com/airspeed-velocity/asv/issues/966
        self.setup()
        with self.versioned_file.stage_version("version2") as g:
            dataset = g["data"]
            assert (
                isinstance(dataset, InMemoryDataset)
                or isinstance(dataset, DatasetWrapper)
                and isinstance(dataset.dataset, InMemoryDataset)
            )
            dataset.resize((10, 10, 10))



from importlib.metadata import version

from .api import VersionedHDF5File
from .replay import delete_version, delete_versions, modify_metadata

__version__ = version(__package__)

__all__ = ["VersionedHDF5File", "delete_version", "delete_versions", "modify_metadata"]



import itertools

import h5py
import numpy as np
import pytest
from numpy.testing import assert_equal

from ..api import VersionedHDF5File
from ..wrappers import InMemoryArrayDataset, InMemoryGroup, InMemorySparseDataset


@pytest.fixture()
def premade_group(h5file):
    group = h5file.create_group("group")
    return InMemoryGroup(group.id)


def test_InMemoryArrayDataset(h5file):
    group = h5file.create_group("group")
    parent = InMemoryGroup(group.id)
    a = np.arange(
        100,
    ).reshape((50, 2))
    dataset = InMemoryArrayDataset("data", a, parent=parent)
    assert dataset.name == "data"
    assert_equal(dataset.array, a)
    assert dataset.attrs == {}
    assert dataset.shape == a.shape
    assert dataset.dtype == a.dtype
    assert dataset.ndim == 2

    # __array__
    assert_equal(np.array(dataset), a)

    # Length of the first axis, matching h5py.Dataset
    assert len(dataset) == dataset.len() == 50

    # Test __iter__
    assert_equal(list(dataset), list(a))

    assert dataset[30, 0] == a[30, 0] == 60
    dataset[30, 0] = 1000
    assert dataset[30, 0] == 1000
    assert dataset.array[30, 0] == 1000

    assert dataset.size == 100


def test_InMemoryArrayDataset_resize(h5file):
    group = h5file.create_group("group")
    parent = InMemoryGroup(group.id)
    a = np.arange(100)
    dataset = InMemoryArrayDataset("data", a, parent=parent)
    dataset.resize((110,))

    assert len(dataset) == 110
    assert_equal(dataset[:100], dataset.array[:100])
    assert_equal(dataset[:100], a)
    assert_equal(dataset[100:], dataset.array[100:])
    assert_equal(dataset[100:], 0)
    assert dataset.shape == dataset.array.shape == (110,)

    a = np.arange(100)
    dataset = InMemoryArrayDataset("data", a, parent=parent)
    dataset.resize((90,))

    assert len(dataset) == 90
    assert_equal(dataset, dataset.array)
    assert_equal(dataset, np.arange(90))
    assert dataset.shape == dataset.array.shape == (90,)


def test_InMemorySparseDataset(h5file):
    group = h5file.create_group("group")
    parent = InMemoryGroup(group.id)
    d = InMemorySparseDataset(
        "data", shape=(1000,), dtype=np.float64, parent=parent, fillvalue=1.0
    )
    assert d.shape == (1000,)
    assert d.name == "data"
    assert d.dtype == np.float64
    assert d.fillvalue == np.float64(1.0)


def test_InMemorySparseDataset_getitem(h5file):
    group = h5file.create_group("group")
    parent = InMemoryGroup(group.id)
    d = InMemorySparseDataset(
        "data", shape=(1000,), dtype=np.float64, parent=parent, fillvalue=1.0
    )
    assert_equal(d[0], 1.0)
    assert_equal(d[:], np.ones((1000,)))
    assert_equal(d[10:20], np.ones((10,)))


@pytest.mark.parametrize(
    "oldshape,newshape",
    itertools.combinations_with_replacement(
        itertools.product(range(5, 25, 5), repeat=3), 2
    ),
)
@pytest.mark.slow
def test_InMemoryArrayDataset_resize_multidimension(oldshape, newshape, h5file):
    # Test semantics against raw HDF5
    a = np.arange(np.prod(oldshape)).reshape(oldshape)

    group = h5file.create_group("group")
    parent = InMemoryGroup(group.id)

    dataset = InMemoryArrayDataset("data", a, parent=parent, fillvalue=-1)
    dataset.resize(newshape)

    h5file.create_dataset(
        "data", data=a, fillvalue=-1, chunks=(10, 10, 10), maxshape=(None, None, None)
    )
    h5file["data"].resize(newshape)
    assert_equal(dataset[()], h5file["data"][()], str(newshape))


def test_group_repr(premade_group):
    """Test that repr(InMemoryGroup) also shows reprs of child objects."""
    foo = premade_group.create_dataset(
        "foo",
        data=np.array([1, 2, 3, 4, 5, np.nan]),
    )
    bar = premade_group.create_dataset(
        "bar",
        data=np.array([1, 2, 3, 4, 5, np.nan]),
    )
    baz = premade_group.create_dataset(
        "baz",
        data=np.array([1, 2, 3, 4, 5, np.nan]),
    )

    result = repr(premade_group)
    assert repr(foo) in result
    assert repr(bar) in result
    assert repr(baz) in result


def test_committed_propagation():
    """Check that InMemoryGroup propagates the '_committed' state to child instances."""
    name = "testname"
    test_data = np.ones((10,))
    f = h5py.File("foo.h5", "w")
    vfile = VersionedHDF5File(f)

    # Commit some data to nested groups
    with vfile.stage_version("version1", "") as group:
        group.create_dataset(f"{name}/key", data=test_data)
        group.create_dataset(f"{name}/val", data=test_data)

    assert vfile["version1"]._committed
    assert vfile["version1"][name]._committed

    with vfile.stage_version("version2") as group:
        key_ds = group[f"{name}/key"]
        val_ds = group[f"{name}/val"]
        val_ds[0] = -1
        key_ds[0] = 0

    assert vfile["version2"]._committed
    assert vfile["version2"][name]._committed



from __future__ import absolute_import, division, print_function, with_statement

import abc
import random
from unittest import TestCase

import numpy as np
import scipy.stats


class TestDatasetPerformanceBase(TestCase, metaclass=abc.ABCMeta):
    """
    Test cases for the most common use cases where we encounter when we write data to HDF5.

    In general all data has multiple columns which are divided into "keys" and "values". The keys
    determine the identity of the row (this stock, this time) and the values are the associated
    values (the price, ...).


    We have two different implementation methods:
    - "sparse": key and value columns are stored as arrays of equal length and to get the i-th
      "row" you read key0[i], key1[i], ..., val0[i], val1[i], ...
    - "dense": key columns are the labels of the axes of the data and the length of the value
      column is the product of the length of the key columns:
      len(val0) == len(key0) * len(key1) * ...
      To get the i-th row you retrieve
        key0[i // len(key1) // len(key2) // ...],
        key1[(i // len(key2) // len(key3) // ...) % len(key1)],
        key2[(i // len(key3) // len(key4) // ...) % len(key2)],
        ...,
        val0[i], val1[i], ...
      TODO: check the math!
    """

    # models
    RECENCTNESS_POWERLAW_SHAPE = 20.0

    def test_mostly_appends_sparse(
        self,
        num_transactions=250,
        filename="test_mostly_appends_sparse",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial = 1000
        num_rows_per_append = 1000

        if deterministic:
            pct_inserts = 0
            pct_deletes = 0
            pct_changes = 0
        else:
            pct_inserts = 5
            pct_deletes = 1
            pct_changes = 5

        num_inserts = 10
        num_deletes = 10
        num_changes = 10

        times = self._write_transactions_sparse(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial,
            num_transactions,
            num_rows_per_append,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes,
            pct_inserts,
            num_inserts,
        )
        return times

    @classmethod
    @abc.abstractmethod
    def _write_transactions_sparse(
        cls,
        name,
        chunk_size,
        compression,
        versions,
        print_transactions,
        num_rows_initial,
        num_transactions,
        num_rows_per_append,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes,
        pct_inserts,
        num_inserts,
    ):
        pass

    @classmethod
    def _get_rand_fn(cls, dtype):
        if dtype == np.dtype("int64"):
            return lambda size=None: np.random.randint(0, int(1e6), size=size)
        elif dtype == np.dtype("float64"):
            return np.random.rand
        else:
            raise ValueError("implement other dtypes")

    @classmethod
    def _modify_dss_sparse(
        cls,
        key0_ds,
        key1_ds,
        val_ds,
        num_rows_per_append,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes,
        pct_inserts,
        num_inserts,
    ):
        ns = set([len(ds) for ds in [key0_ds, key1_ds, val_ds]])
        assert len(ns) == 1
        n = next(iter(ns))
        # change values
        if random.randrange(0, 100) <= pct_changes:
            r_num_chgs = int(np.random.randn() + num_changes)
            rand_fn = cls._get_rand_fn(val_ds.dtype)
            for b in range(r_num_chgs):
                r = random.randrange(0, n)
                val_ds[r] = rand_fn()
        # delete rows
        if random.randrange(0, 100) <= pct_deletes:
            r_num_dels = max(int(np.random.randn() + num_deletes), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE, size=r_num_dels
            )
            rs = np.unique((pdf * n).astype("int64"))
            minr = min(rs)
            n -= len(rs)
            for ds in [key0_ds, key1_ds, val_ds]:
                arr = ds[minr:]
                arr = np.delete(arr, rs - minr)
                ds.resize((n,))
                ds[minr:] = arr
        # insert rows
        if random.randrange(0, 100) <= pct_inserts:
            rand_num_inss = max(int(np.random.randn() + num_inserts), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE,
                size=rand_num_inss,
            )
            rs = np.unique((pdf * n).astype("int64"))
            minr = min(rs)
            n += len(rs)
            for ds in [key0_ds, key1_ds, val_ds]:
                rand_fn = cls._get_rand_fn(ds.dtype)
                arr = ds[minr:]
                arr = np.insert(arr, rs - minr, [rand_fn() for _ in rs])
                ds.resize((n,))
                ds[minr:] = arr
        # append
        rand_num_apps = int(10 * np.random.randn() + num_rows_per_append)
        if rand_num_apps > 0:
            n += rand_num_apps
            for ds in [key0_ds, key1_ds, val_ds]:
                rand_fn = cls._get_rand_fn(ds.dtype)
                ds.resize((n,))
                ds[-rand_num_apps:] = rand_fn(rand_num_apps)

    def test_large_fraction_changes_sparse(
        self,
        num_transactions=250,
        filename="test_large_fraction_changes_sparse",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial = 5000
        num_rows_per_append = 10

        if deterministic:
            pct_inserts = 0
            pct_deletes = 0
            pct_changes = 0
        else:
            pct_inserts = 1
            pct_deletes = 1
            pct_changes = 90

        num_inserts = 10
        num_deletes = 10
        num_changes = 1000

        times = self._write_transactions_sparse(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial,
            num_transactions,
            num_rows_per_append,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes,
            pct_inserts,
            num_inserts,
        )
        return times

    def test_small_fraction_changes_sparse(
        self,
        num_transactions=250,
        filename="test_small_fraction_changes_sparse",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial = 5000
        num_rows_per_append = 10

        if deterministic:
            pct_inserts = 0
            pct_deletes = 0
            pct_changes = 0
        else:
            pct_inserts = 1
            pct_deletes = 1
            pct_changes = 90

        num_inserts = 10
        num_deletes = 10
        num_changes = 10

        times = self._write_transactions_sparse(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial,
            num_transactions,
            num_rows_per_append,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes,
            pct_inserts,
            num_inserts,
        )
        return times

    def test_large_fraction_constant_sparse(
        self,
        num_transactions=250,
        filename="test_large_fraction_constant_sparse",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial = 5000
        num_rows_per_append = 0  # triggers the constant size test (FIXME)

        pct_inserts = 0
        pct_deletes = 0
        pct_changes = 0

        num_inserts = 10
        num_deletes = 10
        num_changes = 1000

        times = self._write_transactions_sparse(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial,
            num_transactions,
            num_rows_per_append,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes,
            pct_inserts,
            num_inserts,
        )
        return times

    def test_mostly_appends_dense(
        self,
        num_transactions=250,
        filename="test_mostly_appends_dense",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial_0 = 30
        num_rows_initial_1 = 30
        num_rows_per_append_0 = 1

        if deterministic:
            pct_inserts = 0
            pct_deletes = 0
            pct_changes = 0
        else:
            pct_inserts = 5
            pct_deletes = 1
            pct_changes = 5

        num_inserts_0 = 1
        num_inserts_1 = 10
        num_deletes_0 = 1
        num_deletes_1 = 1
        num_changes = 10

        times = self._write_transactions_dense(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial_0,
            num_rows_initial_1,
            num_transactions,
            num_rows_per_append_0,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes_0,
            num_deletes_1,
            pct_inserts,
            num_inserts_0,
            num_inserts_1,
        )
        return times

    @classmethod
    @abc.abstractmethod
    def _write_transactions_dense(
        cls,
        name,
        chunk_size,
        compression,
        versions,
        print_transactions,
        num_rows_initial_0,
        num_rows_initial_1,
        num_transactions,
        num_rows_per_append_0,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes_0,
        num_deletes_1,
        pct_inserts,
        num_inserts_0,
        num_inserts_1,
    ):
        pass

    @classmethod
    def _modify_dss_dense(
        cls,
        key0_ds,
        key1_ds,
        val_ds,
        num_rows_per_append_0,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes_0,
        num_deletes_1,
        pct_inserts,
        num_inserts_0,
        num_inserts_1,
    ):
        n_key0 = len(key0_ds)
        n_key1 = len(key1_ds)
        val_shape = val_ds.shape
        assert val_shape == (n_key0, n_key1)
        # change values
        if random.randrange(0, 100) <= pct_changes:
            r_num_chgs = int(np.random.randn() + num_changes)
            for b in range(r_num_chgs):
                r = (random.randrange(0, n_key0), random.randrange(0, n_key1))
                val_ds[r] = np.random.rand()
        # delete rows
        if random.randrange(0, 100) <= pct_deletes:
            # delete from values in two steps

            # 1. delete from key0 and associated vals
            r_num_dels_0 = max(int(np.random.randn() + num_deletes_0), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE, size=r_num_dels_0
            )
            rs_0 = np.unique((pdf * n_key0).astype("int64"))
            minr_0 = min(rs_0)

            n_key0 -= len(rs_0)
            arr_key0 = key0_ds[minr_0:]
            arr_key0 = np.delete(arr_key0, rs_0 - minr_0)
            key0_ds.resize((n_key0,))
            key0_ds[minr_0:] = arr_key0

            arr_val = val_ds[minr_0:, :]
            val_shape = (val_shape[0] - len(rs_0), val_shape[1])
            val_ds.resize(val_shape)
            arr_val = np.delete(arr_val, rs_0 - minr_0, axis=0)
            val_ds[minr_0:, :] = arr_val

            # 2. delete from key1 and associated vals
            r_num_dels_1 = max(int(np.random.randn() + num_deletes_1), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE, size=r_num_dels_1
            )
            rs_1 = np.unique((pdf * n_key1).astype("int64"))
            minr_1 = min(rs_1)

            n_key1 -= len(rs_1)
            arr_key1 = key1_ds[minr_1:]
            arr_key1 = np.delete(arr_key1, rs_1 - minr_1)
            key1_ds.resize((n_key1,))
            key1_ds[minr_1:] = arr_key1

            arr_val = val_ds[:, minr_1:]
            val_shape = (val_shape[0], val_shape[1] - len(rs_1))
            val_ds.resize(val_shape)
            arr_val = np.delete(arr_val, rs_1 - minr_1, axis=1)
            val_ds[:, minr_1:] = arr_val
        # insert rows
        if random.randrange(0, 100) <= pct_inserts:
            # insert into values in two steps

            # 1. insert into key0 and associated vals
            rand_num_inss_0 = max(int(np.random.randn() + num_inserts_0), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE,
                size=rand_num_inss_0,
            )
            rs_0 = np.unique((pdf * n_key0).astype("int64"))
            minr_0 = min(rs_0)

            arr_key0 = key0_ds[minr_0:]
            arr_key0 = np.insert(
                arr_key0, rs_0 - minr_0, np.random.randint(0, int(1e6), size=len(rs_0))
            )
            n_key0 += len(rs_0)
            key0_ds.resize((n_key0,))
            key0_ds[minr_0:] = arr_key0

            arr_val = val_ds[minr_0:, :]
            val_shape = (val_shape[0] + len(rs_0), val_shape[1])
            val_ds.resize(val_shape)
            arr_val = np.insert(
                arr_val, rs_0 - minr_0, np.random.rand(len(rs_0), n_key1), axis=0
            )
            val_ds[minr_0:, :] = arr_val

            # 2. insert into key1 and associated vals
            rand_num_inss_1 = max(int(np.random.randn() + num_inserts_1), 1)
            pdf = scipy.stats.powerlaw.rvs(
                TestDatasetPerformanceBase.RECENCTNESS_POWERLAW_SHAPE,
                size=rand_num_inss_1,
            )
            rs_1 = np.unique((pdf * n_key1).astype("int64"))
            minr_1 = min(rs_1)

            arr_key1 = key1_ds[minr_1:]
            arr_key1 = np.insert(
                arr_key1, rs_1 - minr_1, np.random.randint(0, int(1e6), size=len(rs_1))
            )
            n_key1 += len(rs_1)
            key1_ds.resize((n_key1,))
            key1_ds[minr_1:] = arr_key1

            arr_val = val_ds[:, minr_1:]
            val_shape = (val_shape[0], val_shape[1] + len(rs_1))
            val_ds.resize(val_shape)
            arr_val = np.insert(
                arr_val, rs_1 - minr_1, np.random.rand(n_key0, len(rs_1)), axis=1
            )
            val_ds[:, minr_1:] = arr_val
        # append
        rand_num_apps_0 = int(np.random.randn() + num_rows_per_append_0)
        if rand_num_apps_0 > 0:
            # append to key0 and associated vals
            n_key0 += rand_num_apps_0
            key0_ds.resize((n_key0,))
            key0_ds[-rand_num_apps_0:] = np.random.randint(
                0, int(1e6), size=rand_num_apps_0
            )

            val_shape = (n_key0, n_key1)
            val_ds.resize(val_shape)
            val_ds[-rand_num_apps_0:, :] = np.random.rand(rand_num_apps_0, n_key1)



"""
Public API functions

Everything outside of this file is considered internal API and is subject to
change.
"""

from __future__ import annotations

import datetime
import logging
from contextlib import contextmanager

import h5py
import ndindex
import numpy as np

from .backend import CORRUPT_DATA_VERSIONS, DATA_VERSION, initialize
from .hashtable import Hashtable
from .slicetools import spaceid_to_slice
from .versions import (
    all_versions,
    commit_version,
    create_version_group,
    delete_version,
    get_nth_previous_version,
    get_version_by_timestamp,
    set_current_version,
)
from .wrappers import (
    DatasetWrapper,
    InMemoryArrayDataset,
    InMemoryDataset,
    InMemoryGroup,
)

logger = logging.getLogger(__name__)


class VersionedHDF5File:
    """
    A Versioned HDF5 File

    This is the main entry-point of the library. To use a versioned HDF5 file,
    pass a h5py file to constructor. The methods on the resulting object can
    be used to view and create versions.

    Note that versioned HDF5 files have a special structure and should not be
    modified directly. Also note that once a version is created in the file,
    it should be treated as read-only. Some protections are in place to
    prevent accidental modification, but it is not possible in the HDF5 layer
    to make a dataset or group read-only, so modifications made outside of
    this library could result in breaking things.

    >>> import h5py
    >>> f = h5py.File('file.h5') # doctest: +SKIP
    >>> from versioned_hdf5 import VersionedHDF5File
    >>> file = VersionedHDF5File(f) # doctest: +SKIP

    Access versions using indexing

    >>> version1 = file['version1'] # doctest: +SKIP

    This returns a group containing the datasets for that version.

    To create a new version, use :func:`stage_version`.

    >>> with file.stage_version('version2') as group: # doctest: +SKIP
    ...     group['dataset'] = ... # Modify the group
    ...

    When the context manager exits, the version will be written to the file.

    Finally, use

    >>> file.close() # doctest: +SKIP

    to close the `VersionedHDF5File` object (note that the `h5py` file object
    should be closed separately.)
    """

    def __init__(self, f):
        self.f = f
        if "_version_data" not in f:
            initialize(f)
        else:
            # This is not a new file; check data version identifier for compatibility
            if self.data_version_identifier < DATA_VERSION:
                if self.data_version_identifier in CORRUPT_DATA_VERSIONS:
                    raise ValueError(
                        f"Versioned Hdf5 file {f.filename} has data_version {self.data_version_identifier}, "
                        "which has corrupted hash_tables. "
                        "See https://github.com/deshaw/versioned-hdf5/issues/256 and "
                        "https://github.com/deshaw/versioned-hdf5/issues/288 for details. "
                        "You should recreate the file from scratch. "
                        "In an emergency you could also rebuild the hash tables by calling "
                        f"with h5py.File({f.filename!r}) as f: VersionedHDF5File(f).rebuild_hashtables() and use "
                        f"delete_versions to delete all versions after the upgrade to "
                        f"data_version == {self.data_version_identifier} if you can identify them."
                    )
                if any(self._find_object_dtype_data_groups()):
                    logger.warning(
                        'Detected dtype="O" arrays which are not reused when creating new versions. '
                        "See https://github.com/deshaw/versioned-hdf5/issues/256 for details. "
                        "Rebuilding hash tables for %s is recommended by calling "
                        "with h5py.File(%r) as f: VersionedHDF5File(f).rebuild_object_dtype_hashtables().",
                        f.filename,
                        f.filename,
                    )
                else:
                    if f.mode == "r+":
                        logger.info(
                            "Ugprading data_version to %d, no action required.",
                            DATA_VERSION,
                        )
                        versions = self.f["_version_data"]["versions"]
                        versions.attrs["data_version"] = DATA_VERSION

            elif self.data_version_identifier > DATA_VERSION:
                raise ValueError(
                    f"{f.filename} was written by a later version of versioned-hdf5"
                    f"than what is currently installed. Please update versioned-hdf5."
                )

        self._closed = False
        self._version_cache = {}

    @property
    def _versions(self):
        """Shorthand reference to the versions group of the file."""
        return self.f["_version_data"]["versions"]

    @property
    def _version_data(self):
        return self.f["_version_data"]

    @property
    def closed(self):
        if self._closed:
            return self._closed
        if not self.f.id:
            self._closed = True
        return self._closed

    @property
    def current_version(self):
        """
        The current version.

        The current version is used as the default previous version to
        :func:`stage_version`, and is also used for negative integer version
        indexing (the current version is `self[0]`).
        """
        return self._versions.attrs["current_version"]

    @current_version.setter
    def current_version(self, version_name):
        set_current_version(self.f, version_name)
        self._version_cache.clear()

    @property
    def data_version_identifier(self) -> str:
        """Return the data version identifier.

        Different versions of versioned-hdf5 handle data slightly differently.
        This string affects whether the version of versioned-hdf5 is compatible with the
        given file.

        If no data version attribute is found, it is assumed to be `1`.

        Returns
        -------
        str
            The data version identifier string
        """
        return self.f["_version_data/versions"].attrs.get("data_version", 1)

    @data_version_identifier.setter
    def data_version_identifier(self, version: int):
        """Set the data version identifier for the current file.

        Parameters
        ----------
        version : int
            Version value to write to the file.
        """
        self.f["_version_data/versions"].attrs["data_version"] = version

    def get_version_by_name(self, version):
        if version.startswith("/"):
            raise ValueError(
                "Versions cannot start with '/'. VersionedHDF5File should not be used to access the top-level of an h5py File."
            )

        if version == "":
            version = "__first_version__"

        if version not in self._versions:
            raise KeyError(f"Version {version!r} not found")

        g = self._versions[version]
        if not g.attrs["committed"]:
            raise ValueError(
                "Version groups cannot accessed from the VersionedHDF5File object before they are committed."
            )
        if self.f.file.mode == "r":
            return g
        return InMemoryGroup(g._id, _committed=True)

    def get_version_by_timestamp(self, timestamp, exact=False):
        version = get_version_by_timestamp(self.f, timestamp, exact=exact)
        g = self._versions[version]
        if not g.attrs["committed"]:
            raise ValueError(
                "Version groups cannot accessed from the VersionedHDF5File object before they are committed."
            )
        if self.f.file.mode == "r":
            return g
        return InMemoryGroup(g._id, _committed=True)

    def __getitem__(self, item):
        if self.closed:
            raise ValueError("File is closed")
        if item in self._version_cache:
            # We don't cache version names because those are already cheap to
            # lookup.
            return self._version_cache[item]

        if item is None:
            return self.get_version_by_name(self.current_version)
        elif isinstance(item, str):
            return self.get_version_by_name(item)
        elif isinstance(item, (int, np.integer)):
            if item > 0:
                raise IndexError("Integer version slice must be negative")
            self._version_cache[item] = self.get_version_by_name(
                get_nth_previous_version(self.f, self.current_version, -item)
            )
            return self._version_cache[item]
        elif isinstance(item, (datetime.datetime, np.datetime64)):
            self._version_cache[item] = self.get_version_by_timestamp(item)
            return self._version_cache[item]
        else:
            raise KeyError(f"Don't know how to get the version for {item!r}")

    def __delitem__(self, item):
        """
        Delete a version

        If the version is the current version, the new current version will be
        set to the previous version.
        """
        if not isinstance(item, str):
            raise NotImplementedError("del is only supported for string keys")
        if item not in self._versions:
            raise KeyError(item)
        new_current = (
            self.current_version
            if item != self.current_version
            else self[item].attrs["prev_version"]
        )
        delete_version(self.f, item, new_current)
        self._version_cache.clear()

    def __iter__(self):
        return all_versions(self.f, include_first=False)

    @contextmanager
    def stage_version(
        self,
        version_name: str,
        prev_version=None,
        make_current=True,
        timestamp=None,
    ):
        """
        Return a context manager to stage a new version

        The context manager returns a group, which should be modified in-place
        to build the new version. When the context manager exits, the new
        version will be written into the file.

        `version_name` should be the name for the version.

        `prev_version` should be the previous version which this version is
        based on. The group returned by the context manager will mirror this
        previous version. If it is `None` (the default), the previous
        version will be the current version. If it is `''`, there will be no
        previous version.

        If `make_current` is `True` (the default), the new version will be set
        as the current version. The current version is used as the default
        `prev_version` for any future `stage_version` call.

        `timestamp` may be a datetime.datetime or np.datetime64 timestamp for
        the version. Note that datetime.datetime timestamps must be in the UTC
        timezone (np.datetime64 timestamps are not timezone aware and are
        assumed to be UTC). If `timestamp` is `None` (the default) the current
        time when the context manager exits is used. When passing in a manual
        timestamp, be aware that no consistency checks are made to ensure that
        version timestamps are linear or not duplicated.

        """
        if self.closed:
            raise ValueError("File is closed")
        old_current = self.current_version
        group = create_version_group(self.f, version_name, prev_version=prev_version)

        try:
            yield group
            group.close()
            commit_version(
                group,
                group.datasets(),
                make_current=make_current,
                chunks=group.chunks,
                compression=group.compression,
                compression_opts=group.compression_opts,
                timestamp=timestamp,
            )
        except:
            delete_version(self.f, version_name, old_current)
            raise
        finally:
            self._version_cache.clear()

    def close(self):
        """
        Make sure the VersionedHDF5File object is no longer reachable.
        """
        if not self._closed:
            del self.f
            self._closed = True

    def __repr__(self):
        """
        Prints friendly status information.

        These messages are intended to be similar to h5py messages.
        """
        if self.closed:
            return "<Closed VersionedHDF5File>"
        else:
            return (
                f'<VersionedHDF5File object "{self.f.filename}" (mode'
                f" {self.f.mode})>"
            )

    def rebuild_hashtables(self):
        """Delete and rebuild *all* existing hashtables for the raw datasets."""

        data_groups = self._find_all_data_groups()

        self._rebuild_hashtables(data_groups)

    def _find_all_data_groups(self):
        # Find all data groups excluding '/_version_data/versions'
        data_groups = []
        for name, group in self.f["_version_data"].items():
            if name != "versions":
                data_groups.extend(self._find_data_groups(group))
        return data_groups

    def _rebuild_hashtables(self, data_groups):
        """Rebuild the hashtables in data_groups."""
        for group in data_groups:
            del self.f[group.name]["hash_table"]
            Hashtable.from_versions_traverse(self.f, group.name)

    def _find_data_groups(self, node: h5py.Group) -> list[h5py.Group]:
        """Find all groups containing datasets that are descendents of the given node.

        Parameters
        ----------
        node : h5py.Group
            Node under which groups containing datasets live.

        Returns
        -------
        list[h5py.Group]
            List of groups which hold versioned-hdf5 datasets. Each group should
            contain a h5py.Dataset named 'raw_data'.
        """
        items = []
        if isinstance(node, h5py.Group):
            if "raw_data" in node:
                items.append(node)
            else:
                for item in node:
                    items.extend(self._find_data_groups(self.f[f"{node.name}/{item}"]))
        return items

    def _find_object_dtype_data_groups(self):
        """Find all data groups with dtype='O'."""

        # Find all data groups excluding '/_version_data/versions' of dtype='O'
        for data_group in self._find_all_data_groups():
            if data_group["raw_data"].dtype.kind == "O":
                yield data_group

    def rebuild_object_dtype_hashtables(self):
        """Find all dtype='O' data groups and rebuild their hashtables."""
        logger.info(
            'Rebuilding hash tables for dtype="O" datasets in %s.', self.f.filename
        )
        self._rebuild_hashtables(self._find_object_dtype_data_groups())
        self.f["_version_data"]["versions"].attrs["data_version"] = DATA_VERSION

    def get_diff(
        self, name: str, version1: str, version2: str
    ) -> dict[tuple[slice, ...], tuple[np.ndarray, ...]]:
        """Compute the difference between two versions of a dataset.

        Parameters
        ----------
        name : str
            Name of the dataset
        version1 : str
            First version to compare
        version2 : str
            Second version to compare

        Returns
        -------
        dict[tuple[slice, ...], tuple[np.ndarray, ...]]
            A dictionary where the keys are slices that changed from version1 to
            version2, the the values are tuples containing

                (data_in_version1, data_in_version2)
        """
        if self.closed:
            raise ValueError("File is not open.")

        data1 = self[version1][name]
        data2 = self[version2][name]

        if isinstance(data1, DatasetWrapper):
            data1 = data1.dataset
        if isinstance(data2, DatasetWrapper):
            data2 = data2.dataset

        if isinstance(data1, (InMemoryDataset, InMemoryArrayDataset)) or isinstance(
            data2, (InMemoryDataset, InMemoryArrayDataset)
        ):
            raise ValueError(
                "Versions have not yet been committed to the file. "
                "Please close and reopen the file before calling get_diff()."
            )

        return self._diff_data(
            version1,
            version2,
            name,
            self._get_chunks(name, version1),
            self._get_chunks(name, version2),
        )

    def _get_chunks(
        self,
        name: str,
        version: str,
    ) -> dict[ndindex.Tuple, ndindex.Tuple]:
        """Get the Dict which maps virtual dataset sources to raw data slices.

        Parameters
        ----------
        name : str
            Name of the dataset
        version : str
            Version of the dataset for which the data map is to be retrieved

        Returns
        -------
        dict[ndindex.Tuple, ndindex.Tuple]
            Mapping between the dataset's virtual slices and the slices of the
            underlying raw data
        """
        slices = {}
        for vs in self[version][name].virtual_sources():
            src = spaceid_to_slice(vs.src_space)
            vir = spaceid_to_slice(vs.vspace)
            slices[vir] = src

        return slices

    def _diff_data(
        self,
        v1: str,
        v2: str,
        name: str,
        sd1: dict[ndindex.Tuple, ndindex.Tuple],
        sd2: dict[ndindex.Tuple, ndindex.Tuple],
    ) -> dict[ndindex.Tuple, tuple[np.ndarray, ...]]:
        """Compute the difference between two versions of a dataset.

        Parameters
        ----------
        v1 : str
            Version of a dataset
        v2 : str
            Version of a different dataset
        name : str
            Name of the dataset
        sd1 : dict[ndindex.Tuple, ndindex.Tuple]
            Dict which maps virtual dataset slices to source (raw) dataset slices
        sd2 : dict[ndindex.Tuple, ndindex.Tuple]
            Dict which maps virtual dataset slices to source (raw) dataset slices

        Returns
        -------
        dict[ndindex.Tuple, tuple[np.ndarray, ...]]:
            Dict of {changed_virtual_slice: (data_v1, data_v2)} which maps slices of the
            virtual datasets which were modified between the versions to a tuple
            containing the corresponding raw ndarray that was modified. The first
            element of the tuple is the data at version `v1`, and the second element
            is the data at version `v2`. If data was added or removed, that element
            of the tuple will be None. Example return value:

            diff = {
                # Data was deleted
                Tuple(Slice(0, 1024, 1)): (array([1, ... 1]), None),

                # Data was added
                Tuple(Slice(1024, 2048, 1)): (None, array([1, ... 1])),

                # Data was changed
                Tuple(Slice(2048, 3072, 1)): (array([1, ... 1]), array([2, ..., 2])),
            }

        """
        diff = {}
        for vir2, src2 in sd2.items():
            if vir2 not in sd1:
                # New data was written
                diff[vir2] = (None, self[v2][name][vir2.raw])
            elif src2 != sd1[vir2]:
                # Data was modified
                diff[vir2] = (self[v1][name][vir2.raw], self[v2][name][vir2.raw])

        for vir1, src1 in sd1.items():
            if vir1 not in sd2:
                # Data was deleted
                diff[vir1] = (self[v1][name][vir1.raw], None)

        return diff

    @property
    def versions(self) -> list[str]:
        """Return the names of the version groups in the file.

        This should return the same as calling
        ``versioned_hdf5.versions.all_versions(self.f, include_first=False)``.

        Returns
        -------
        list[str]
            The names of versions in the file; order is arbitrary
        """
        return [v for v in self._versions if "__first_version__" not in v]
